
#include "Application.hpp"

#include <exception>
#include <string>
#include <algorithm> // swap
#include <stdexcept>
#include <cstdlib>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/none.hpp>
#include <boost/optional.hpp>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include "Settings.hpp"
#include "application-constants.hpp"
#include "HistoricModel.hpp"
#include "Canvas.hpp"
#include "Observers.hpp"
#include "Rotators.hpp"
#include "geometry.hpp"
#include "util/ScopeGuard.hpp"
#include "util/TimeoutTimer.hpp"
#include "util/glademm.hpp"
#include "util/bfstream.hpp"
#include "util/gettext.hpp"

#include <boost/filesystem/operations.hpp>

namespace fs = boost::filesystem;

namespace {

std::string get_env (std::string const & s)
{
  if (char const * const r = std::getenv(s.c_str())) return r;
  else throw std::runtime_error(_("Environment variable not set: ") + s);
}

template <typename F>
struct exception_catcher
  // Top level wrapper for things that exceptions must not fall through; currently used for signal handlers and main. Regarding exception_catcher's use with menu item signal handlers, note that execution of f is assumed to be transactional (if an exception occurs, the program state must not have been altered), basic error safe (no resources must leak), and obviously nonfatal.
{
  F f;

  void operator() ()
  {
    std::string m;
    try { f(); return; }
    catch (std::exception const & e) { m = e.what(); }
    catch (Glib::Exception const & e) { m = e.what(); }
    catch (...) { m = _("An unknown error occurred."); }

    Gtk::MessageDialog(m, false /* no markup */, Gtk::MESSAGE_ERROR,
                       Gtk::BUTTONS_OK, true /* modal */).run();
  }
};

template <typename F>
exception_catcher<F> make_exception_catcher (F const f)
{
  exception_catcher<F> const r = { f };
  return r;
}

} // unnamed namespace

namespace M3D {

namespace {

float const timeout = 2; // mostly for grid-generation

char const * const timeout_error_msg = _("Could not generate grid with the requested properties.");

POD_DEFINE(GameData, (( MinedProperty, mines   ))\
                     (( FlaggedProperty, flags ))\
                     (( OpenedProperty, opened )))

boost::shared_ptr<HistoricModel> makeModel (Settings const & s, TimeoutTimer const & t,
                                            HistoricModel const * const ex_m = 0,
                                            Settings const * const ex_s = 0,
                                            GameData const * const gd = 0)
{
  assert(bool(ex_m) == bool(ex_s)); // either both set or both unset
  assert(!bool(ex_m) || !bool(gd)); // we can't process both an existing model and gamedata

  // Since the grid is stateless, there's no point in re-using any existing model's grid (other than efficiency, but that's insignificant here).

  boost::shared_ptr<HistoricModel> r;

  if (ex_m &&
      s->modelSettings.gridSettings == (*ex_s)->modelSettings.gridSettings &&
      s->modelSettings.mineRatio    == (*ex_s)->modelSettings.mineRatio)
  {
    assert(hasFieldVisualizationSelector(*ex_m));

    // Other settings (autohelpers) can be applied cleanly on the existing model, so just copy it.
    r.reset(new HistoricModel(*ex_m));

    r->get<FlagObserver>().setCycleThroughRisky(s->cycleThroughRisky);

    // It is important that we do all necessary observer removal before any necessary observer addition, because otherwise adding an observer might cause it to interact with another observer that was scheduled for removal.

    if (s->modelSettings.autoOpen != (*ex_s)->modelSettings.autoOpen)
    {
      if (s->modelSettings.autoOpen) removeMinimalAutoOpener(*r);
      else removeNormalAutoOpener(*r);
    }

    if (s->modelSettings.autoFlag != (*ex_s)->modelSettings.autoFlag && !s->modelSettings.autoFlag)
      removeAutoFlagger(*r);

    if (s->modelSettings.autoOpen != (*ex_s)->modelSettings.autoOpen)
    {
      if (s->modelSettings.autoOpen) addNormalAutoOpener(*r);
      else addMinimalAutoOpener(*r);
    }

    if (s->modelSettings.autoFlag != (*ex_s)->modelSettings.autoFlag && s->modelSettings.autoFlag)
      addAutoFlagger(*r);

    if (s->subtractFlagged != (*ex_s)->subtractFlagged)
    {
      removeFieldVisualizationSelector(*r);
      addFieldVisualizationSelector(*r, s->subtractFlagged);
    }
  }
  else
  {
    r.reset(new HistoricModel(ObservedModel(BareModel(s->modelSettings.gridSettings))));

    if (gd) addMineObserver(*r, s->modelSettings.mineRatio, s->modelSettings.maxAdjacentMines, gd->mines);
    else addMineObserver(*r, s->modelSettings.mineRatio, s->modelSettings.maxAdjacentMines, t);

    if (gd) addOpenObserver(*r, gd->opened);
    else addOpenObserver(*r);

    if (gd) r->add(FlagObserver(*r, s->cycleThroughRisky, &(gd->flags)));
    else r->add(FlagObserver(*r, s->cycleThroughRisky));

    addFieldVisualizationSelector(*r, s->subtractFlagged);

    if (s->modelSettings.autoOpen) addNormalAutoOpener(*r);
    else addMinimalAutoOpener(*r);

    if (s->modelSettings.autoFlag) addAutoFlagger(*r);

    addAutoStarter(*r, t);

    if (!gd) restartUntilPlaying(*r, t);

    r->clearPast();
  }

  return r;
}

boost::optional<fs::path> save_dialog (std::string const & message, fs::path const & default_path)
{
  Gtk::FileChooserDialog d (message, Gtk::FILE_CHOOSER_ACTION_SAVE);
  d.set_current_folder(default_path.native_directory_string());
  d.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  d.add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);
  boost::optional<fs::path> r;
  if (d.run() == Gtk::RESPONSE_OK) r = fs::path(d.get_filename(), fs::native);
  return r;
}

} // unnamed namespace

class Application : boost::noncopyable
/*
- is responsible for creating/handling the menu's
- owns the main model as well as the Canvas, and binds them together
*/
{
  boost::optional<fs::path> gameFile;
  Settings settings; // Invariant: These apply to model and canvas.
  boost::shared_ptr<HistoricModel> model;
  Canvas canvas; // Invariant: This displays model.
  Glib::RefPtr<Gnome::Glade::Xml> const gladeXml;

  void menu_game_new ()
  {
    restart(TimeoutTimer(timeout));
    gameFile = boost::none;
  }

  void save (fs::path const & p) const
  {
    bofstream f (p);
    boost::archive::text_oarchive a (f);
    GameData const gd = { mined(*model), flagLevel(*model), opened(*model) };
    a << settings->modelSettings << gd;
  }

  void menu_game_save ()
  {
    if (gameFile) save(*gameFile);
    else menu_game_save_as();
  }

  void menu_game_save_as ()
  {
    Gtk::FileChooserDialog d (_("Save Game As"), Gtk::FILE_CHOOSER_ACTION_SAVE);
    d.set_current_folder((app_user_path()/"saves").native_directory_string());
    d.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
    d.add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);
    if (d.run() != Gtk::RESPONSE_OK) return;
    fs::path const p (d.get_filename(), fs::native);
    save(p);
    gameFile = p;
  }

  void menu_game_open ()
  {
    Gtk::FileChooserDialog d (_("Open Game"), Gtk::FILE_CHOOSER_ACTION_OPEN);
    d.set_current_folder((app_user_path()/"saves").native_directory_string());
    d.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
    d.add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);
    if (d.run() != Gtk::RESPONSE_OK) return;
    fs::path const p (d.get_filename(), fs::native);
    load(p);
    gameFile = p;
  }

  void menu_game_preferences ()
  {
    if (boost::optional<Settings> const newsettings = dialog(settings, gladeXml))
    {
      set(*newsettings, TimeoutTimer(timeout));
      saveToFile(app_user_path()/"settings", settings);
    }
  }

  void menu_game_undo () { undo(*model); canvas.invalidate(); }
  void menu_game_redo () { redo(*model); canvas.invalidate(); }

  void menu_game_selected ()
  {
    get<Gtk::MenuItem>(gladeXml, "menu_game_undo").set_sensitive(canUndo(*model));
    get<Gtk::MenuItem>(gladeXml, "menu_game_redo").set_sensitive(canRedo(*model));
  }

  void menu_game_deselected ()
  {
    get<Gtk::MenuItem>(gladeXml, "menu_game_undo").set_sensitive(true);
    get<Gtk::MenuItem>(gladeXml, "menu_game_redo").set_sensitive(true);

    // In general, checking whether undo/redo are available passively (that is, by being notified when they become available or unavailable) is hard. It is much, much easier to actively perform a check when the operation is requested. For example, rather than trying to keep the sensitivity of the undo/redo menu items in-sync with the actual availability of the operations, we simply set the sensitivity when the menu is opened. This works, but there's a catch: the keyboard accelerator shortcuts corresponding to the menu items simply do not work when those menu items aren't sensitive. Rather than falling back to passive availability notification, we work around this problem by simply having the menu items being sensitive when the menu isn't selected. Note that this doesn't cause problems when the shortcuts are used while the operation is unavailable, because availability is checked before the operation is executed anyway.
  }

  void menu_help_help ()
  {
    std::string const txt = str(boost::format(_(
      "The object of %1% is to locate all hidden mines in a three-dimensional field grid where each field may or may not cover a mine.\n\n"
      "Initially, all fields, regardless of whether they cover a mine, are covered. You can open fields by clicking on them. If you open a field that covers a mine, the mine explodes and you lose. If however the field does not cover a mine, a digit will apear on the location of the field, indicating how many adjacent fields cover mines.\n\n"
      "By analyzing groups of these numbers you can try to figure out which covered fields are likely to cover mines. If you think you have found a field covering a mine, you can place a flag on it by clicking on it with the secondary mouse button. This saves you from having to remember these locations yourself, and guides the support algorithms if you are using them.\n\n"
      "Once you uncover all fields except those that cover mines, you win.\n\n"
      "Here are a few tips in case you're still having trouble:\n"
      "- get used to Minesweeper 2D first, then give %1% another try;\n"
      "- use a low mine ratio to get a quick impression of what the game is all about."
    )) % APP_NAME);

    Gtk::MessageDialog(txt, false /* no markup */, Gtk::MESSAGE_INFO, Gtk::BUTTONS_OK, true /* modal */).run();
  }

  void menu_help_about ()
  {
    std::string const txt = str(boost::format(_(
      "%1%\n\n"
      "version %2%\n\n"
      "by %3%\n\n"
      "This software has been released into the Public Domain."
    )) % APP_NAME % APP_VERSION % "Eelis van der Weegen");

    Gtk::MessageDialog(txt, false /* no markup */, Gtk::MESSAGE_INFO, Gtk::BUTTONS_OK, true /* modal */).run();

/*
    Gtk::Dialog & d = get<Gtk::Dialog>(gladeXml, "aboutDialog");
    d.show_all();
    ScopeGuard const hider (boost::bind(&Gtk::Dialog::hide_all, &d));
    d.run();
*/
  }

  void restart (TimeoutTimer const & t)
  {
    makeModel(settings, t).swap(model);
    canvas.set(model);
  }

  void load (fs::path const & p)
  {
    bifstream f (p);
    boost::archive::text_iarchive a (f);
    GameData gd;
    Settings::base_type u = settings.base();

    a >> u.modelSettings >> gd;
    Settings const s (u);

    TimeoutTimer const t (timeout);

    makeModel(s, t, 0, 0, &gd).swap(model);

    try
    {
      settings = s;
      canvas.set(model);
      canvas.set(s->canvasSettings);
    }
    catch (...) { assert(!"Application class invariants violated."); }
      // There basically is no solution for this, as Canvas simply cannot be made strong exception safe.
  }

  void set (Settings const & s, TimeoutTimer const & t)
  {
    makeModel(s, t, model.get(), &settings).swap(model);

    try
    {
      settings = s;
      canvas.set(model);
      canvas.set(s->canvasSettings);
    }
    catch (...) { assert(!"Application class invariants violated."); }
      // There basically is no solution for this, as Canvas simply cannot be made strong exception safe.
  }

  public:

    Application ()
     : settings (loadSettingsFrom(app_user_path()/"settings")),
       model (makeModel(settings, TimeoutTimer(timeout))),
       canvas (model, settings->canvasSettings),
       gladeXml (Gnome::Glade::Xml::create(APP_RESOURCE_PATH"/gui.glade"))
    {
      assert(gladeXml); // create() should have thrown an exception if anything was wrong

      get<Gtk::MenuItem>(gladeXml, "menu_game").signal_deselect().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_deselected, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_game").signal_select().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_selected, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_game_quit").signal_activate().connect(
        &Gtk::Main::quit);
      get<Gtk::MenuItem>(gladeXml, "menu_game_new").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_new, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_game_open").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_open, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_game_save").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_save, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_game_save_as").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_save_as, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_game_undo").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_undo, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_game_redo").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_redo, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_game_preferences").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_game_preferences, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_help_help").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_help_help, this)));
      get<Gtk::MenuItem>(gladeXml, "menu_help_about").signal_activate().connect(
        make_exception_catcher(boost::bind(&Application::menu_help_about, this)));

      get<Gtk::VBox>(gladeXml, "topVbox").pack_start(canvas);
      Gtk::Window & w = get<Gtk::Window>(gladeXml, "mainWindow");
      w.set_reallocate_redraws(true);
      w.show_all();
      Gtk::Main::run(w);
    };
};

fs::path app_user_path () { return fs::path(get_env("HOME") + "/." APP_ID, fs::native); }

} // namespace M3D

void caught_main (int argc, char * argv [])
{
  setlocale(LC_ALL, "");
  bindtextdomain(APP_ID, APP_LOCALE_PATH);
  textdomain(APP_ID);

  Gtk::Main const kit (argc, argv);
  if (!Gtk::GL::init_check(argc, argv)) throw std::runtime_error(_("Could not initialize OpenGL."));
  if (argc > 1) throw std::runtime_error(_("Unrecognized command line parameter passed: ") + std::string(argv[1]));

  srand(time(0));

  fs::path const p = M3D::app_user_path()/"settings";
  if (!fs::exists(p))
  {
    fs::create_directory(M3D::app_user_path());
    fs::create_directory(M3D::app_user_path()/"saves");
    fs::create_directory(M3D::app_user_path()/"skins");
    fs::copy_file(APP_RESOURCE_PATH"/default_settings", p);
  }

  M3D::Application();
}

POD_DEFINE_SERIALIZATION(M3D::GameData)

int main (int argc, char * argv [])
{
  make_exception_catcher(boost::bind(&caught_main, argc, argv))();
}


