
#include "Observers.hpp"

#include "Operations.hpp"

#include "util/foreach.hpp"
#include "util/visitation.hpp"

namespace M3D {

namespace {

// AutoStarter

struct AutoStarter : ObservedModel::ObserverBase<AutoStarter, 10 /* order, sync: see notes.txt */>
{
  void onOperation (ObservedModel & model, Operation const & o)
  {
    if (RestartOperation const * const p = dynamic_cast<RestartOperation const *>(&o))
      if (!openInitial(model)) restart(model, p->t);
        // we keep using the existing TimeoutTimer because this is still a single RestartOperation from the user's perspective
  }

  static bool openInitial (ObservedModel & model) // returns whether successful
  {
    // Todo: AutoStarting a 0 field in the (hardly visible) center of a model is pointless. It would be better to prefer fields at or near the edge of the model.

    std::vector<Field> v;
    v.reserve(num_vertices(model.grid()));

    // First let's check whether perhaps there's already an opened 0-field (if so we're done). In the process, add all candidates to v.

    BOOST_FOREACH(Field const f, vertices(model.grid()))
      if (adjacentMines(model, f) == 0 && !mined(model)[f])
      {
        if (opened(model)[f])
          return true;
        v.push_back(f);
      }

    if (v.empty()) return false;

    std::random_shuffle(v.begin(), v.end()); // Efficiency: This is overkill, we just need one random element.

    open(v.front(), model);

    return true;
  }
};

// MinimalAutoOpener

class MinimalAutoOpener : public ObservedModel::ObserverBase<MinimalAutoOpener, 10 /* order, sync: see notes.txt */>,
                          Visitor<OpenOperation const>
{
  ObservedModel * model;

  void consider (Field const f) const
  {
    if (opened(*model)[f] && adjacentMines(*model, f) == 0)
      BOOST_FOREACH(Field const g, adjacent_vertices(f, model->grid()))
        open(g, *model);
  }

  void visit (OpenOperation const & o) const { consider(o.field); }

  void onOperation (ObservedModel & m, Operation const & o) { model = &m; o.accept(*this); }

  public:

    MinimalAutoOpener (ObservedModel & m) : model (&m) {}

    void considerAll () const { BOOST_FOREACH(Field const f, vertices(model->grid())) consider(f); }
};

// NormalAutoOpener

class NormalAutoOpener : public ObservedModel::ObserverBase<NormalAutoOpener, 10 /* order, sync: see notes.txt */>,
                         Visitor<FlagOperation const>,
                         Visitor<RestartOperation const>,
                         Visitor<OpenOperation const>
{
  ObservedModel * model;

  void visit (FlagOperation const & o) const
  {
    if (flagLevel(*model)[o.field] != flagged) return;
    BOOST_FOREACH(Field const f, adjacent_vertices(o.field, model->grid())) consider(f);
  }

  void visit (RestartOperation const &) const { considerAll(); }
  void visit (OpenOperation const & o) const { consider(o.field); }

  void onOperation (ObservedModel & m, Operation const & o)
  {
    model = &m;
    o.accept(*this);
  }

  void consider (Field const f) const
  {
    if (!opened(*model)[f]) return;

    unsigned int flags = 0;
    BOOST_FOREACH(Field const g, adjacent_vertices(f, model->grid()))
      if (flagLevel(*model)[g] == flagged) ++flags;
    if (flags >= adjacentMines(*model, f))
      BOOST_FOREACH(Field const g, adjacent_vertices(f, model->grid()))
        if (flagLevel(*model)[g] != flagged) open(g, *model);
  }

  public:

    void considerAll () const { BOOST_FOREACH(Field const f, vertices(model->grid())) consider(f); }

    NormalAutoOpener (ObservedModel & m) : model (&m) {}
};

// AutoFlagger

class AutoFlagger : public ObservedModel::ObserverBase<AutoFlagger, 10 /* order, sync: see notes.txt */>,
                    Visitor<RestartOperation const>,
                    Visitor<OpenOperation const>,
                    Visitor<FlagOperation const>
{
  ObservedModel * model;

  void consider (Field const f) const
  {
    if (!opened(*model)[f]) return;

    unsigned int c = 0;

    BOOST_FOREACH(Field const g, adjacent_vertices(f, model->grid()))
      if (!opened(*model)[g]) ++c;

    if (c == adjacentMines(*model, f))
      BOOST_FOREACH(Field const g, adjacent_vertices(f, model->grid()))
        flag(g, *model);
  }

  void consider_around (Field const f) const
  {
    BOOST_FOREACH(Field const g, adjacent_vertices(f, model->grid()))
      consider(g);
  }

  void visit (RestartOperation const &) const { considerAll(); }
  void visit (FlagOperation const & o) const { consider_around(o.field); }
  void visit (OpenOperation const & o) const { if (o.opened) consider_around(o.field); }

  void onOperation (ObservedModel & m, Operation const & o) { model = &m; o.accept(*this); }

  public:

    void considerAll () const { BOOST_FOREACH(Field const f, vertices(model->grid())) consider(f); }

    AutoFlagger (ObservedModel & m) : model (&m) {}
};

} // unnamed namespace

void addMinimalAutoOpener (ObservedModel & model)
{
  model.add(MinimalAutoOpener(model));
  model.get<MinimalAutoOpener>().considerAll();
}

void addNormalAutoOpener (ObservedModel & model)
{
  model.add(NormalAutoOpener(model));
  model.get<NormalAutoOpener>().considerAll();
}

void addAutoFlagger (ObservedModel & model)
{
  model.add(AutoFlagger(model));
  model.get<AutoFlagger>().considerAll();
}

void addAutoStarter (ObservedModel & model, TimeoutTimer const & t)
{
  model.add(AutoStarter());
  if (!AutoStarter::openInitial(model)) restart(model, t);
}

void removeMinimalAutoOpener (ObservedModel & model) { model.remove<MinimalAutoOpener>(); }
void removeNormalAutoOpener (ObservedModel & model) { model.remove<NormalAutoOpener>(); }
void removeAutoFlagger (ObservedModel & model) { model.remove<AutoFlagger>(); }
void removeAutoStarter (ObservedModel & model) { model.remove<AutoStarter>(); }

} // namespace M3D
