/*

Invariant: GL_MATRIX_MODE is GL_MODELVIEW

  Note that this invariant refers to the global OpenGL state, which means that maintaining it is by definition a responsibility of all code that modifies that state. This is why as much OpenGL state modification must be done here and not elsewhere: the more fragmented the OpenGL state modification code is, the harder it is to ensure these invariants.

Mouse field selection

  The traditional way to do mouse selection in OpenGL is by using glRenderMode GL_SELECT with a gluPickMatrix. Previous versions of M3D used this technique along with some extra machinery to deal with partially transparent textures. This worked fine until I tried to implement highlighting the same way. The problem is that the whole GL_SELECT procedure is relatively costly, and highlighting would require us to do it every time a mouse move event occurs. While performance was acceptable for tiny grid sizes, it was unacceptable for bigger grid sizes. Another approach was needed.

  The fundamental problem with the traditional approach described above is that it requires lots of (additional) geometric and rendering operations for every mouse move operation. It occured to me that ideally we would like to have some sort of data structure that directly maps viewport coordinates to field indices, which we would update every time the model state or rotation changes. That way, all that would be needed on a mouse move event was a lookup in that data structure. It then occured to me that we can use OpenGL's off-screen buffer for this purpose. The trick is to draw the scene as usual (from the same perspective, etc), but leave out the sprites, and instead of drawing the fields with their assigned colors draw them completely opaque and without any lighting, in a color that encodes their field index. Then, mapping the cursor location to a field is as simple as reading a pixel from the offscreen buffer and decoding the color value to yield the field index (if any).

  If OpenGL were 100% color precise and reliable, encoding/decoding could be as easy as just chopping the field index up into a 3-byte tuple corresponding to the RGB color channels. Unfortunately, OpenGL is not 100% color precise and reliable. For example, anti-aliasing is not always disable-able, and when it's not it messes up the field<->color mapping at field borders. I've come up with two strategies to avoid this problem.

  The first strategy was based on a clever encoding/decoding scheme that reduced the chance of mismatches at field borders by only mapping to color codes in which all color channel values were 0 modulo 25. This strategy seemed to work fine, but was dumped because the chance of mismatches was greater than with the second strategy.

  The second strategy, which is currently used by M3D, is based on the idea that anti-aliasing is very pixel-oriented. The chance that two or more adjacent anti-aliased pixels are of the same color is extremely small. Therefore, if the cursor points to a pixel which has a few adjacent pixels of the same color, we conclude that it must be because they are part of the same field surface (or the background, which is a special case). On the other hand, if the cursor points to a pixel whose adjacent pixels are of another color, we conclude that we're at a field border and abort the decoding. This strategy has a far smaller chance of mismatches, but is slightly less accurate cursor-wise and slightly more costly since we have to read a few more pixels.

  Note that we currently assume that OpenGL /is/ 100% color precise for large opaque non-aliased surfaces. If it turns out that this assumption doesn't hold on some systems, we'll have to use some approximation/rounding/spreading encoding/decoding scheme (note that the second second strategy discussed above is much more suitable for this than the first).

Display list

  OpenGL features so-called "display lists". Display lists have strengths and weaknesses. In particular, executing a display list is efficient, while compiling a display lists is expensive. Canvas uses the display list feature when it is sensible to do so:
  - If any form of highlighting is enabled, any mouse movements that causes another field to be highlighted would cause display list recompilation, which is too expensive.
  - Non-opaque fields and textures need to be Z-ordered manually, which rules out putting them in a display list.

  Efficiency: Making a large display list takes a lot of time. When we change the visualization of fields by clicking on them we do not want to wait that long for the result, so either we should compile+run the list at the same time, or do a "normal" drawing without any display list, followed by the list creation.

  Invariant: The set of fields in fieldSorting and the set of fields contained in displayList do not intersect, and combined form the set of all visible fields (and sprites).

Todo: Demand double buffering and other OpenGL features we really need, and verify that we get them.

Fragile: The update/invalidate code was written in a pretty ad-hoc fashion, merely guided by what appeared to work.

Todo: A lot of procedures and subsystems here need thorough documentation.

Todo: Do we really have to acquire and validate the gldrawable on each expose event? Can't we just do it once?

*/

#include "Canvas.hpp"
#include "HistoricModel.hpp"
#include "application-constants.hpp"
#include "util/foreach.hpp"
#include "util/GLutils.hpp"
#include "util/ScopeGuard.hpp"
#include <boost/none.hpp>
#include <boost/bind.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/implicit_cast.hpp>
#include <librsvg/rsvg.h>
#include <stdexcept>

#include <GL/glu.h>

namespace fs = boost::filesystem;

namespace M3D {

namespace
{
  template <typename T, template <typename> class L>
  std::vector<T> vector (L<T> const & l) { return l; }
    // matching against boost::assign_detail::generic_list<T> would be nicer, but we're supposed to remain unaware of detail

  unsigned int const DEFAULT_DRAGBOX_WIDTH = 4,
                     DEFAULT_DRAGBOX_HEIGHT = 4;

  template <typename float_t>
  boost::numeric::ublas::c_matrix<float_t, 4, 4> rotation_to_matrix (boost::math::quaternion<float_t> const & q)
  {
    float_t const
      a  = q.R_component_1(),
      b  = q.R_component_2(),
      c  = q.R_component_3(),
      d  = q.R_component_4(),
      ad = a * d, ac = a * c, ab = a * b, aa = a * a,
      bd = b * d, bc = b * c, bb = b * b,
      cd = c * d, cc = c * c,
      dd = d * d,
      norme_carre = aa + bb + cc + dd;

    if (norme_carre <= std::numeric_limits<float_t>::epsilon())
      throw std::underflow_error("Argument is too small.");

    float_t const m [] =
      {     (aa + bb - cc - dd) / norme_carre, 2 * (bc - ad          ) / norme_carre, 2 * (ac + bd          ) / norme_carre, 0,
        2 * (ad + bc)           / norme_carre,     (aa - bb + cc - dd) / norme_carre, 2 * (cd - ab          ) / norme_carre, 0,
        2 * (bd - ac)           / norme_carre, 2 * (ab + cd          ) / norme_carre,     (aa - bb - cc + dd) / norme_carre, 0,
        0, 0, 0, 1 };

    boost::numeric::ublas::c_matrix<float_t, 4, 4> r;
    std::copy(m, m + 16, r.data());
    return r;
  }

  bool highlightable (FieldVisualization const fv)
  {
    return fv == neutral_vis || fv == flag_vis || fv == potentialMine_vis;
  }

  Glib::RefPtr<Gdk::GL::Config> createGLconfig (Gdk::GL::ConfigMode const cm)
  {
    if (Glib::RefPtr<Gdk::GL::Config> const r = Gdk::GL::Config::create(cm)) return r;
    else throw std::runtime_error("Could not create OpenGL configuration.");
      // Todo: Doesn't Gdk::GL::Config::create throw exceptions itself?
  }

  struct PairFirstLess
  {
    template <typename T, typename Y>
    bool operator() (std::pair<T, Y> const & a, std::pair<T, Y> const & b) const
    {
      return a.first < b.first;
    }
  };

  Glib::RefPtr<Gdk::Pixbuf> load_texture (fs::path const & skin, std::string const & name, unsigned int const size)
  {
    // Todo: Better error handling.

    RsvgHandle * h = rsvg_handle_new();
    if (!h) throw std::runtime_error("could not create rsvg handle");
    ScopeGuard const freeer (boost::bind(rsvg_handle_free, h));
    GError * e = 0;

    std::stringstream s; s <<
      "<?xml version='1.0' encoding='UTF-8'?>"
      "<!DOCTYPE svg PUBLIC '-//W3C//DTD SVG 20010904//EN' 'http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd'>"
      "<svg width='" << size << "' height='" << size << "' xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>"
      " <g transform='scale(" << size << ")'><use xlink:href='" << skin.native_file_string() << "#" << name << "'/></g>"
      "</svg>";

    // Todo: Properly escape the resource dir and name strings.

    std::istreambuf_iterator<char> beg (s), end;
    std::vector<unsigned char> const svg (beg, end);
    if (rsvg_handle_write(h, &svg[0], svg.size(), &e) == FALSE)
      throw std::runtime_error("rsvg_handle_write failed");
    if (rsvg_handle_close(h, &e) == FALSE)
      throw std::runtime_error("rsvg_handle_close failed");
    GdkPixbuf * const gpb = rsvg_handle_get_pixbuf(h);
    if (!gpb) std::runtime_error("could not get pixbuf from rsvg");

    Glib::RefPtr<Gdk::Pixbuf> const r = Glib::wrap(gpb)->flip(false /* vertical */);

    assert(r->get_bits_per_sample() == 8);
    assert(r->get_has_alpha());
    assert(r->get_n_channels() == 4);
    assert(r->get_rowstride() == r->get_width() * r->get_n_channels());
    assert(r->get_width() == int(size) && r->get_height() == int(size));

    return r;
  }

  // See the "Mouse field selection" section at the top of this file for the purpose of encodeField/decodeField. Sync: these must be eachother's inverse.

  RGBcolor Canvas::encodeField (Field const f) const
  {
    return RGBcolor((f & 0x0000ff) >> 0,
                    (f & 0x00ff00) >> 8,
                    (f & 0xff0000) >> 16);
  }

  Field Canvas::decodeField (RGBcolor const & c) const
  {
    return c.r() | (c.g() << 8) | (c.b() << 16);
  }

} // unnamed namespace

void Canvas::draw (FieldVisualization const fv)
{
  fieldDrawers[fv].draw(*(model->gridSettings().fieldShape), (fv > 0 && fv < 10 ? settings.digitSize : settings.solidFieldSize));
}

void Canvas::onOperation (ObservedModel &, Operation const &)
{
  fieldSortingOutdated = displayListOutdated = true;
}

bool Canvas::highlighted (Field const f) const
{
  if (!pointedToField || !highlightable(visualization(*model)[f])) return false;

  switch (settings.highlightStyle)
  {
    case HighlightStyle::single: return f == *pointedToField;
    case HighlightStyle::adjacent:
    {
      FieldShape const & fs = *(model->gridSettings().fieldShape);

      return f != *pointedToField &&
              fs.adjacent(model->grid()[f].coords,
                          model->grid()[*pointedToField].coords, model->gridSettings().lessDetection);
    }
    default: return false;
  }
}

void Canvas::invalidate () { get_window()->invalidate_rect(get_allocation(), false); }

void Canvas::setProjectionMatrix () const
{
  glMatrixModeSwitch ms (GL_PROJECTION);
  glLoadIdentity();

  GLint vp [4];
  glGetIntegerv(GL_VIEWPORT, vp);
  assert(vp[0] == 0 && vp[1] == 0);
  if (vp[2] <= 0 || vp[3] <= 0) return;

  float const
    nearPlane      = cameraDistance - sphereRadius,
    farPlane       = cameraDistance + sphereRadius,
    eyeTan         = std::tan(settings.fieldOfVision / 2) * nearPlane,
    ratio          = float(vp[2]) / vp[3],
    xo             = eyeTan * std::max(ratio, 1.0f),
    yo             = eyeTan * std::max(1 / ratio, 1.0f);

  if (glGetError() != GL_NO_ERROR) throw std::runtime_error("f");

  glFrustum(-xo, xo, -yo, yo, nearPlane - 0.01, farPlane);
}

void Canvas::paintColorMap ()
{
  glClearColor(1, 1, 1, 1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  glTranslatef(0, 0, -cameraDistance);
  glMultMatrixf(rotation_to_matrix(rotator->getRotation()).data());
  glDisable(GL_LIGHTING);
  glBindTexture(GL_TEXTURE_2D, 0);

  FieldShape const & fieldShape = *(model->gridSettings().fieldShape);

  BOOST_FOREACH(Field const f, vertices(model->grid()))
  {
    FieldVisualization const fv = visualization(*model)[f];

    if (highlightable(fv))
    {
      glStackedMatrix sm;
      FieldCoordinates const & c = model->grid()[f].coords;
      glTranslatef(c(0), c(1), c(2));
      glScalef(settings.solidFieldSize);
      glColor3ubv(encodeField(f).elems);
      // Efficiency: It's been said that having the glBegin and glEnd outside the loop can actually decrease performance. Look into it.
      glBegin(GL_QUADS);
        BOOST_FOREACH(FieldShape::Side const& side, fieldShape.sides())
          BOOST_FOREACH(FieldShape::Vertex const & v, side.quad) glVertex3fv(v.data());
      glEnd();
    }
  }

  colorMapInBuffer = true;
}

void Canvas::onDrag (double const from_x, double const from_y,
                     double const to_x, double const to_y)
{
  float const w = get_width(),
              h = get_height();
  assert(w > 0 && h > 0); // dragging when there's no client area makes no sense
  rotator->onDrag(from_x / w - 0.5,
                  to_x   / w - 0.5,
                  from_y / h - 0.5,
                  to_y   / h - 0.5);
  fieldSortingOutdated = true;
}

void Canvas::onSphereRadiusChange ()
{
  float newSphereRadius = 0;

  BOOST_FOREACH(Field const f, vertices(model->grid()))
  {
    boost::numeric::ublas::c_vector<float, 3> const t = model->grid()[f].coords;
    newSphereRadius = std::max(newSphereRadius, float(std::sqrt(inner_prod(t, t))));
  }

  newSphereRadius += fieldDrawers[neutral_vis].radius(*(model->gridSettings().fieldShape), settings.solidFieldSize);

  assert(newSphereRadius > 0);

  cameraDistance = newSphereRadius / std::sin(settings.fieldOfVision / 2);
  sphereRadius = newSphereRadius;

  setProjectionMatrix();
}

bool Canvas::on_button_press_event (GdkEventButton * const e)
{
  if (e->button == 1) dragStartPoint = prevMousePosition = std::make_pair(e->x, e->y);
  return false;
}

bool Canvas::on_button_release_event (GdkEventButton * const e)
{
   // Todo: Describe what it means for dragStartPoint to be set/unset.
  if (e->button == 1 && dragStartPoint) onClick(e->x, e->y, true);
  else if (e->button == 3) onClick(e->x, e->y, false);
  return false;
}

void Canvas::onClick (double const x, double const y, bool const primary)
{
  if (playState(*model) == playing)
    if (boost::optional<Field> const f = getFieldUnder(x, y))
      onFieldClick(*f, primary);
}

void Canvas::onFieldClick (Field const f, bool const primary)
{
  if (!primary)
  {
    model->setHistoryMark();
    if (!M3D::opened(*model)[f]) M3D::cycleFlagLevel(f, *model);
    fieldSortingOutdated = displayListOutdated = true;
  }
  else if (!opened(*model)[f] && flagLevel(*model)[f] != flagged)
  {
    model->setHistoryMark();
    M3D::open(f, *model);
    fieldSortingOutdated = displayListOutdated = true;
  }
  else return;

  invalidate();

  if (playState(*model) == won)
    Gtk::MessageDialog("Congratulations, you successfully located all mines!", false /* no markup */,
                       Gtk::MESSAGE_INFO, Gtk::BUTTONS_OK, true /* modal */).run();
}

boost::optional<Field> Canvas::getFieldUnder (double const x, double const y)
{
  if (x < 0 || y < 0 || x >= get_width() || y >= get_height()) return boost::none;

  if (!colorMapInBuffer) paintColorMap();

  // See the "Mouse field selection" section at the top of this file for a discussion of what follows.

  RGBAcolor pixels [4];
  glReadPixels(int(x), int(get_height() - y), 2, 2, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
    // We really only need GL_RGB, but reading GL_RGBA values is easier since 4 channels align better (since 4 is a power of 2).

  if (pixels[0] != pixels[1] ||
      pixels[0] != pixels[2] ||
      pixels[0] != pixels[3]) return boost::none;

  Field const f = decodeField(RGBcolor(pixels[0]));

  return (f == 0xffffff ? boost::none : boost::optional<Field>(f));
}

bool Canvas::on_motion_notify_event(GdkEventMotion * const e)
{
  int dragBoxWidth = DEFAULT_DRAGBOX_WIDTH,
      dragBoxHeight = DEFAULT_DRAGBOX_HEIGHT; // Todo: get system defaults, and check circularly.

  bool repaint = false;

  if (e->state & GDK_BUTTON1_MASK &&
      (!dragStartPoint ||
        std::abs(dragStartPoint->first - e->x) > dragBoxWidth / 2 ||
        std::abs(dragStartPoint->second - e->y) > dragBoxHeight / 2))
  {
    dragStartPoint = boost::none;
    onDrag(prevMousePosition->first, prevMousePosition->second, e->x, e->y);
    repaint = true;
  }

  if (!prevMousePosition || prevMousePosition->first != e->x || prevMousePosition->second != e->y) // moved
  {
    prevMousePosition = std::make_pair(e->x, e->y);

    if (settings.highlightStyle != HighlightStyle::none && !(e->state & GDK_BUTTON1_MASK))
      // We don't change the highlighted field(s) when we're dragging because it looks messy.
    {
      boost::optional<Field> const prevPointedToField = pointedToField;
      pointedToField = getFieldUnder(e->x, e->y);
      if (pointedToField != prevPointedToField) repaint = true;
    }
  }

  if (repaint) invalidate();

  return false;
}

void Canvas::on_realize ()
{
  Gtk::DrawingArea::on_realize();
    // This causes on_configure_event to be triggered. However, by that time we haven't done any OpenGL initialization yet (in particular, the field drawers haven't been loaded yet), so it won't be able to do anything useful. Now, on one of my development workstations on_configure_event was automatically called a second time later on, but on another it wasn't. That is why we explicitly call it ourselves at the end of this function. Note that calling it several times is completely harmless.

  Glib::RefPtr<Gdk::GL::Drawable> gldrawable = get_gl_drawable();

  if (!get_gl_drawable()->gl_begin(get_gl_context())) throw std::runtime_error("could not begin drawing on GL drawable");

  model->add(forwardingObserver<0>(boost::bind(&Canvas::onOperation, this, _1, _2)));

  glEnable(GL_LIGHT1);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_NORMALIZE);
  glEnable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE); // see notes.txt
  glLightfv(GL_LIGHT1, GL_DIFFUSE, &vector(boost::assign::list_of(0.4f)(0.4f)(0.4f)(1.0f))[0]);

  glLightiv(GL_LIGHT1, GL_POSITION, &vector(boost::assign::list_of(1)(1)(1)(0))[0]);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, &vector(boost::assign::list_of(0.7f)(0.7f)(0.7f)(1.0f))[0]);
  glShadeModel(GL_FLAT);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glMatrixMode(GL_MODELVIEW); // This is the "idle" mode. Functions that change it must set it back to GL_MODELVIEW when they're done (or fail).

  displayList = glGenLists(1);

  loadSkin();

  add_events(Gdk::POINTER_MOTION_MASK |
             Gdk::BUTTON_PRESS_MASK   |
             Gdk::BUTTON_RELEASE_MASK );

  on_configure_event(0);
}

void Canvas::loadSkin ()
{
  backgroundColor = *reinterpret_cast<RGBAcolor const *>(load_texture(settings.skin, "background", 1 /* size */)->get_pixels());

  // sync: fieldvisualization definition in Observers.hpp
  fieldDrawers.clear();
  fieldDrawers.push_back(nullFieldDrawer);
  for (int digit = 1; digit != 10; ++digit)
    fieldDrawers.push_back(spriteFieldDrawer(load_texture(settings.skin, "digit" +
     boost::lexical_cast<std::string>(digit), settings.textureSize), 0.5));
  fieldDrawers.push_back( solidFieldDrawer(load_texture(settings.skin, "neutral"      , settings.textureSize)));
  fieldDrawers.push_back( solidFieldDrawer(load_texture(settings.skin, "flag"         , settings.textureSize)));
  fieldDrawers.push_back( solidFieldDrawer(load_texture(settings.skin, "potentialMine", settings.textureSize)));
  fieldDrawers.push_back(spriteFieldDrawer(load_texture(settings.skin, "mine"         , settings.textureSize), 0.6));
  fieldDrawers.push_back(spriteFieldDrawer(load_texture(settings.skin, "crossedMine"  , settings.textureSize), 0.6));
  fieldDrawers.push_back(spriteFieldDrawer(load_texture(settings.skin, "explosion"    , settings.textureSize), 1));

  highlighted_neutral_drawer      .reset(new FieldDrawer(solidFieldDrawer(load_texture(settings.skin, "neutral-highlighted",       settings.textureSize))));
  highlighted_flag_drawer         .reset(new FieldDrawer(solidFieldDrawer(load_texture(settings.skin, "flag-highlighted",          settings.textureSize))));
  highlighted_potentialMine_drawer.reset(new FieldDrawer(solidFieldDrawer(load_texture(settings.skin, "potentialMine-highlighted", settings.textureSize))));
}

bool Canvas::on_configure_event (GdkEventConfigure *)
{
  if (fieldDrawers.empty()) return true; // See the documentation in on_realize.

  if (!get_gl_drawable()->gl_begin(get_gl_context()))
    throw std::runtime_error("could not begin context on OpenGL drawable");
  int const w = get_width(), h = get_height();
  assert(w >= 0 && h >= 0);
  glViewport(0, 0, w, h);

  onSphereRadiusChange();

  return true;
}

bool Canvas::on_expose_event (GdkEventExpose *)
{
  Glib::RefPtr<Gdk::GL::Drawable> const gldrawable = get_gl_drawable();

  if (!gldrawable->is_double_buffered()) throw std::runtime_error("GL drawable not double buffered");
    // Todo: This is a strange place to check for double bufferedness. Can't we do it in on_configure_event or something?

  if (!gldrawable->gl_begin(get_gl_context()))
    throw std::runtime_error("could not begin context on OpenGL drawable");

  glClearColor(backgroundColor[0] / 255.0,
               backgroundColor[1] / 255.0,
               backgroundColor[2] / 255.0,
               backgroundColor[3] / 255.0);

  if (displayListOutdated)
  {
    glNewList(displayList, GL_COMPILE);

    try
    {
      if (settings.highlightStyle == HighlightStyle::none)
        BOOST_FOREACH(Field const f, vertices(model->grid()))
          if (!highlighted(f))
            if (FieldVisualization const fv = visualization(*model)[f])
              if (fieldDrawers[fv].suitableForDisplayList())
              {
                glStackedMatrix sm;
                FieldCoordinates const & c = model->grid()[f].coords;
                glTranslatef(c(0), c(1), c(2));
                draw(fv);
              }
    }
    catch (...) { assert(!"unrecoverable exception occured when building display list"); }

    glEndList();

    displayListOutdated = false;
  }

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();

  glTranslatef(0, 0, -cameraDistance);

  glMultMatrixf(rotation_to_matrix(rotator->getRotation()).data());

  if (fieldSortingOutdated)
  {
    fieldSorting.reserve(num_vertices(model->grid()));
    fieldSorting.clear();
    GLmatrix const n = getModelViewMatrix();

    BOOST_FOREACH(Field const f, vertices(model->grid()))
    {
      if (FieldVisualization const fv = visualization(*model)[f])
        if (settings.highlightStyle != HighlightStyle::none || !fieldDrawers[fv].suitableForDisplayList())
        {
          GLvector w = model->grid()[f].coords;
          w.resize(4); w(3) = 1;
          w = prod(w, n);
          w.resize(3);
          fieldSorting.push_back(std::make_pair(inner_prod(w, w), f));
        }
    }

    std::sort(fieldSorting.begin(), fieldSorting.end(), PairFirstLess());

    fieldSortingOutdated = false;
  }

  glCallList(displayList);

  for (FieldSorting::reverse_iterator i = fieldSorting.rbegin(), e = fieldSorting.rend(); i != e; ++i)
  {
    FieldCoordinates const & c = model->grid()[i->second].coords;
    glStackedMatrix sm;
    glTranslatef(c(0), c(1), c(2));

    Field const f = i->second;
    FieldVisualization const vis = visualization(*model)[f];
    FieldShape const & shape = *(model->gridSettings().fieldShape);

    if (highlighted(f))
    {
      float const mag = settings.solidFieldSize;

      switch (vis)
      {
        case neutral_vis: highlighted_neutral_drawer->draw(shape, mag); break;
        case flag_vis: highlighted_flag_drawer->draw(shape, mag); break;
        case potentialMine_vis: highlighted_potentialMine_drawer->draw(shape, mag); break;
        default: assert(false);
      }
    }
    else draw(vis);
  }

  colorMapInBuffer = false;

  gldrawable->swap_buffers();

  return true;
}

void Canvas::set (boost::shared_ptr<HistoricModel> const m)
{
  model = m;
  model->add(forwardingObserver<0>(boost::bind(&Canvas::onOperation, this, _1, _2)));
  onSphereRadiusChange();
  fieldSortingOutdated = displayListOutdated = true;
  invalidate();
}


void Canvas::set (CanvasSettings const & s)
{
  rotator.reset(s.rotation->create(s.rotationSpeed, rotator.get()));
  settings = s;
  loadSkin();
  fieldSortingOutdated = displayListOutdated = true;
  onSphereRadiusChange(); // because solidFieldSize may have changed
  invalidate();
}

Canvas::Canvas (boost::shared_ptr<HistoricModel> const m, CanvasSettings const & s)
: Gtk::GL::DrawingArea (createGLconfig(Gdk::GL::MODE_RGB   |
                                       Gdk::GL::MODE_DEPTH |
                                       Gdk::GL::MODE_DOUBLE)),
  model (m), settings (s),
  rotator (s.rotation->create(s.rotationSpeed)),
  colorMapInBuffer (false),
  displayListOutdated (true),
  fieldSortingOutdated (true)
{
  // We can't do any actual OpenGL initialization here, because the OpenGL context isn't available until on_realize.

  // Todo: If exceptions are thrown here, the display list is not released (since that happens in the destructor).
  add_events(Gdk::VISIBILITY_NOTIFY_MASK);
  set_size_request(s.width, s.height);
}

Canvas::~Canvas ()
{
  glDeleteLists(displayList, 1);
    // Note that we blatantly assume that displayList has been set by on_realize by the time we destruct.
}

} // namespace M3D
