
#include "FieldDrawers.hpp"
#include <vector>
#include <boost/scoped_ptr.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/implicit_cast.hpp>
#include "geometry.hpp"
#include "util/foreach.hpp"
#include "util/VirtuallyDestructed.hpp"
#include "util/GLutils.hpp"
#include "util/color.hpp"

namespace M3D {

namespace
{
  template <typename L>
  struct LookupSorter
  // Stores an indexable object by value, so pass an iterator or pointer rather than the container itself.
  {
    L l;
    explicit LookupSorter (L l_) : l (l_) {}
    template <typename T> bool operator() (T a, T b) const { return l[a] < l[b]; }
  };

  template <typename L>
  LookupSorter<L> lookupSorter (L const l) { return LookupSorter<L>(l); }

  void set_texture (Glib::RefPtr<Gdk::Pixbuf> const & p)
    // assumes p is nonempty RGB or RGBA and strided without padding
  {
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glTexImage2D(GL_TEXTURE_2D, 0 /* base image level */, p->get_n_channels(), p->get_width(), p->get_height(),
                  0 /* border width */, (p->get_has_alpha() ? GL_RGBA : GL_RGB), GL_UNSIGNED_BYTE, p->get_pixels());
  }

  bool is_opaque (RGBAcolor const & c) { return c[3] == 255; }
  bool is_not_opaque (RGBAcolor const & c) { return !is_opaque(c); }

} // unnamed namespace

struct FieldDrawer::Implementation : boost::noncopyable, VirtuallyDestructed
{
  virtual void draw (FieldShape const &, float mag) const = 0;
  virtual bool suitableForDisplayList () const = 0;
  virtual float radius (FieldShape const &, float mag) const = 0;
};

FieldDrawer::FieldDrawer (Implementation const * const i) : pimpl (i) {}
FieldDrawer::FieldDrawer (FieldDrawer const & d) : pimpl (d.pimpl) {}
FieldDrawer & FieldDrawer::operator= (FieldDrawer const & d) { pimpl = d.pimpl; return *this; }
FieldDrawer::~FieldDrawer () {}

void FieldDrawer::draw (FieldShape const & s, float const mag) const { pimpl->draw(s, mag); }
bool FieldDrawer::suitableForDisplayList () const { return pimpl->suitableForDisplayList(); }
float FieldDrawer::radius (FieldShape const & fs, float const mag) const { return pimpl->radius(fs, mag); }


namespace {

class NullImplementation : public FieldDrawer::Implementation
{
  void draw (FieldShape const &, float /* magnification */) const {}
  bool suitableForDisplayList () const { return true; }
  float radius (FieldShape const &, float /* magnification */) const { return 0; }

  public:

    NullImplementation () {}
};

class SpriteImplementation : public FieldDrawer::Implementation
{
  boost::scoped_ptr<Texture> const texture;
  float const magnification;

  bool suitableForDisplayList () const { return false; }

  float radius (FieldShape const &, float /* mag */) const { return 0; }

  void draw (FieldShape const & fieldShape, float const mag) const
  {
    // Note that at this point we have two magnification values: "mag" and "magnification". The former corresponds to the solidFieldSize setting and is passed to all field drawers by canvas when it's drawing the grid. The latter was bound to this particular field drawer at the moment is was constructed, and is primarily used to make the explosion sprite bigger than the digit and mine sprites.

    texture->bind();

    glDisable(GL_LIGHTING);

    GLmatrix const m = getModelViewMatrix();

    GLfloat const q = magnification * mag * fieldShape.spriteSize();

    GLfloat const n [] =
      GL_MATRIX(q, 0, 0, m(3, 0),
                0, q, 0, m(3, 1),
                0, 0, q, m(3, 2) + 0.1,
                m(0, 3), m(1, 3), m(2, 3), m(3, 3));

      /* This matrix:
      - turns to face the camera (by overwriting the existing upper-left 3x3 fields)
      - magnifies (q)
      - moves a little to the camera to get in front of the grid lines (if there are any) (m[14])
      */

    glLoadMatrixf(n);
    glColor4f(1, 1, 1, 1);

    glBegin(GL_QUADS);
      glNormal3i(0, 0, -1);
      glTexCoord2i(0, 0); glVertex2f(-0.5, -0.5);
      glTexCoord2i(1, 0); glVertex2f( 0.5, -0.5);
      glTexCoord2i(1, 1); glVertex2f( 0.5,  0.5);
      glTexCoord2i(0, 1); glVertex2f(-0.5,  0.5);
    glEnd();
  }

  public:

    SpriteImplementation (Glib::RefPtr<Gdk::Pixbuf> const & r, float const m)
      // assumes r is nonempty RGBA and strided without padding
      : texture (new Texture), magnification (m)
    {
      texture->bind();
      set_texture(r);
    }
};


class SingleColorSideDrawer
{
  RGBAcolor const color;

  public:

    void before_draw () const
    {
      glBindTexture(GL_TEXTURE_2D, 0);
      glColor4ubv(color.elems);
    }

    bool suitableForDisplayList () const { return is_opaque(color); }

    explicit SingleColorSideDrawer (RGBAcolor const & c) : color (c) {}
};

class TextureSideDrawer
{
  boost::shared_ptr<Texture> const texture;

  bool const opaque;

  static bool is_opaque (Glib::RefPtr<Gdk::Pixbuf> const & p)
  {
    RGBAcolor const * i = reinterpret_cast<RGBAcolor const *>(p->get_pixels()),
              * const e = i + p->get_width() * p->get_height();
    return std::find_if(i, e, is_not_opaque) == e;
  }

  public:

    explicit TextureSideDrawer (Glib::RefPtr<Gdk::Pixbuf> const & p)
      : texture (new Texture), opaque (is_opaque(p))
    {
      texture->bind();
      set_texture(p);
    }

    bool suitableForDisplayList () const { return opaque; }

    void before_draw () const
    {
      texture->bind();
      glColor4f(1, 1, 1, 1);
    }
};

template <typename SideDrawer>
class SolidImplementation : public FieldDrawer::Implementation //, noncopyable (through FieldDrawer::Implementation)
{
  SideDrawer sideDrawer;

  bool suitableForDisplayList () const { return sideDrawer.suitableForDisplayList(); }

  float radius (FieldShape const & fs, float const mag) const
  {
    float q = 0;
    typedef boost::numeric::ublas::c_vector<float, 3> V;
    BOOST_FOREACH(FieldShape::Side const & side, fs.sides())
      BOOST_FOREACH(V t, side.quad)
        q = std::max(q, inner_prod(t, t));
    assert(q != 0);
    return sqrt(q) * mag;
  }

  void draw (FieldShape const & fieldShape, float const mag) const
  {
    glEnable(GL_LIGHTING);
    glStackedMatrix sm;
    glScalef(mag);

    sideDrawer.before_draw();

    // Efficiency: It's been said that having the glBegin and glEnd outside the loop can actually decrease performance. Look into it.
    glBegin(GL_QUADS);
      BOOST_FOREACH(FieldShape::Side const& side, fieldShape.sides())
      {
        glNormal3fv(side.normal.data());

        glTexCoord2i(0, 0); glVertex3fv(side.quad[0].data());
        glTexCoord2i(1, 0); glVertex3fv(side.quad[1].data());
        glTexCoord2i(1, 1); glVertex3fv(side.quad[2].data());
        glTexCoord2i(0, 1); glVertex3fv(side.quad[3].data());
          // Todo: Is calling glTextCoord2i really harmless when no texture is bound?
      }
    glEnd();
  }

  public:

    explicit SolidImplementation (SideDrawer const & sd) : sideDrawer(sd) {}
};

} // unnamed namespace

FieldDrawer solidFieldDrawer (Glib::RefPtr<Gdk::Pixbuf> const r)
  // assumes r is nonempty RGBA and strided without padding
{
  RGBAcolor const * const b = reinterpret_cast<RGBAcolor const *>(r->get_pixels()),
                  * const e = b + r->get_width() * r->get_height();
  return FieldDrawer(std::find_if(b, e, boost::lambda::_1 != *b) != e
                      ? boost::implicit_cast<FieldDrawer::Implementation *>(new SolidImplementation<TextureSideDrawer>(TextureSideDrawer(r)))
                      : boost::implicit_cast<FieldDrawer::Implementation *>(new SolidImplementation<SingleColorSideDrawer>(SingleColorSideDrawer(*b))));
}

// Todo: Consider unpimpling FieldDrawer since Canvas no longer passes them around by value anyway.

FieldDrawer spriteFieldDrawer (Glib::RefPtr<Gdk::Pixbuf> const r, float const m)
{
  return FieldDrawer(new SpriteImplementation(r, m));
}

FieldDrawer const nullFieldDrawer (new NullImplementation);

} // namespace M3D
