
// The public interface of this unit consists of the three functions visualization(), addFieldVisualizationSelector(), and removeFieldVisualizationSelector() which are declared in Observers.hpp and defined at the end of this file.

#include "Observers.hpp"
#include "Operations.hpp"

#include "util/foreach.hpp"

namespace
{
  template <typename T>
  inline T proper_subtract (T const & a, T const & b)
  {
    return ((b > a) ? 0 : (a - b));
  }
}

namespace M3D {

namespace {

bool hide (ObservedModel const & model, Field const f)
{
  if (!opened(model)[f] || mined(model)[f])
    return false;

  BOOST_FOREACH(Field const f, adjacent_vertices(f, model.grid()))
    if (!opened(model)[f] && flagLevel(model)[f] != flagged)
      return false;

  return true;
}

class FieldVisualizationSelector : public ObservedModel::ObserverBase<FieldVisualizationSelector,
                                                                      10 /* order, sync: see notes.txt */>,
                                   Visitor<OpenOperation const>,
                                   Visitor<FlagOperation const>,
                                   Visitor<RestartOperation const>
{
  ObservedModel const * model;
  bool subtractFlagged;
  Visualizations visualizations;
  PlayState playState_; // we keep track of the playState so we can detect when it changes.

  void onOperation (ObservedModel & m, Operation const & o)
  {
    model = &m;

    PlayState const p = playState(*model);
    if (p != playState_)
    {
      playState_ = p;
      considerAll();
      return;
    }

    o.accept(*this);
  }

  unsigned int detectCount (Field const f) const
  {
    unsigned int r = adjacentMines(*model, f);

    if (!subtractFlagged) return r;

    unsigned int c = 0;

    BOOST_FOREACH(Field const g, adjacent_vertices(f, model->grid()))
      if (!opened(*model)[g] && flagLevel(*model)[g] == flagged) ++c;

    return proper_subtract(r, c);
  }

  void init ()
  {
    visualizations.resize(num_vertices(model->grid()));
    std::fill(visualizations.begin(), visualizations.end(), empty_vis);
    considerAll();
  }

  void consider (Field const f)
  {
    FieldVisualization & s = visualizations[f];

    if (playState_ == won)
      s = (mined(*model)[f] ? flag_vis : empty_vis);
    else if (hide(*model, f))
      s = empty_vis;
    else if (playState_ == playing)
    {
      if (opened(*model)[f])
        s = FieldVisualization(detectCount(f));
      else
        switch (flagLevel(*model)[f])
        {
          case flagged: s = flag_vis; break;
          case risky: s = potentialMine_vis; break;
          case none: s = neutral_vis; break;
          default: assert(!"invalid flag level value");
        }
    }
    else if (playState_ == lost)
    {
      if (opened(*model)[f])
        s = mined(*model)[f] ? fatalMine_vis : FieldVisualization(detectCount(f));
      else
        switch (flagLevel(*model)[f])
        {
          case flagged: s = mined(*model)[f] ? flag_vis : falseFlag_vis; break;
          case risky: s = mined(*model)[f] ? mine_vis : potentialMine_vis; break;
          case none: s = mined(*model)[f] ? mine_vis : neutral_vis; break;
          default: assert(!"invalid flag level value");
        }
    }
    else assert(!"invalid play state value");
  }

  void considerAll ()
  {
    BOOST_FOREACH(Field const f, vertices(model->grid())) consider(f);
  }

  void visit (OpenOperation const & o)
  {
    consider(o.field);
    BOOST_FOREACH(Field const f, adjacent_vertices(o.field, model->grid())) consider(f);
  }

  void visit (FlagOperation const & o)
  {
    consider(o.field);
    BOOST_FOREACH(Field const f, adjacent_vertices(o.field, model->grid())) consider(f);
  }

  void visit (RestartOperation const &) { init(); }

  friend Visualizations const & M3D::visualization (ObservedModel const &);

  public:

    FieldVisualizationSelector (ObservedModel const & m, bool const sf)
      : model (&m), subtractFlagged (sf), playState_ (playState(*model))
    {
      init();
    }
};

} // unnamed namespace

Visualizations const & visualization (ObservedModel const & model)
{
  return model.get<FieldVisualizationSelector>().visualizations;
}

void addFieldVisualizationSelector (ObservedModel & model, bool const subtractFlagged)
{
  model.add(FieldVisualizationSelector(model, subtractFlagged));
}

bool hasFieldVisualizationSelector (ObservedModel const & model)
{
  return model.getp<FieldVisualizationSelector>();
}

void removeFieldVisualizationSelector (ObservedModel & model)
{
  model.remove<FieldVisualizationSelector>();
}

} // namespace M3D

