
// We can't just store all operations that are apply()'d in our 'past' stack, because that would ignore important operation conflict filtering performed by ObservedModel, leaving us with an incorrect historic account. Instead, we register a ForwardingObserver which will store the (inverses of the) operations when they are actually applied by ObservedModel.

// Since inverting an operation can involve saving field state information (like FlagOperation::inverse() does for example), it must be done _before_ the operation is applied.

#include "HistoricModel.hpp"

#include <cassert>

#include <boost/bind.hpp>
#include <boost/lambda/lambda.hpp>

#include "util/ScopeGuard.hpp"

namespace
{
  template <typename T, typename C>
  void clear (std::stack<T, C> & s) { s = std::stack<T, C>(); }

  template <typename To, typename From>
  struct static_caster
  {
    typedef From & argument_type;
    typedef To & result_type;
    To & operator() (From & r) const { return static_cast<To &>(r); }
  };
}

namespace M3D {

void HistoricModel::onOperation (Operation const & o) // called by forwardingObserver added in constructor
{
  if (!recording) return;

  CloningPtr<Operation> const i (o.inverse(*this));
  if (i.get())
  {
    if (past.empty()) past.push(OperationGroup());
    past.top().push(i);
  }
  else clear(past);
}

void HistoricModel::apply (Operation const & o)
{
  if (!recording) return;
  clear(future);
  BaseModel::apply(o);
}

void HistoricModel::clearPast ()
{
  assert(recording);
  clear(past);
}

void HistoricModel::historyStep (bool const forward)
{
  if (!canHistoryStep(forward)) return;

  assert(recording);
  recording = false;
  ScopeGuard const sg (boost::lambda::var(recording) = true);

  std::stack<OperationGroup> * const a = (forward ? &future : &past),
                             * const b = (forward ? &past : &future);
  OperationGroup inverses;

  while (!a->top().empty())
  {
    CloningPtr<Operation> const i (a->top().top()->inverse(*this));
    BaseModel::apply(*(a->top().top()));
      // All operations stored in 'future' and 'past' are inverses, so if we can't inverse one of their elements it is not symetrically invertible.
    assert(i.get() && "operation not symmetrically invertible");
    inverses.push(i);
    a->top().pop();
  }

  b->push(inverses);
  a->pop();
}

HistoricModel::HistoricModel (BaseModel const & m)
  : BaseModel (m), recording (true)
{
  add(forwardingObserver<-20 /* order, see notes.txt */>
   (boost::bind(&HistoricModel::onOperation, boost::bind(static_caster<HistoricModel, ObservedModel>(), _1), _2)));
}

void HistoricModel::setHistoryMark ()
{
  assert(recording);
  past.push(OperationGroup());
}

bool HistoricModel::canHistoryStep (bool const forward) const
{
  return !(forward ? future : past).empty();
}

} // namespace M3D
