

#include "Observers.hpp"
#include "Operations.hpp"

#include <algorithm> // find, count

#include "util/foreach.hpp"

namespace
{
  template <typename I, typename T>
  bool includes (I const b, I const e, T const & t)
  {
    return std::find(b, e, t) != e;
  }
}

namespace M3D {

namespace {

template <typename Graph>
std::set<typename boost::graph_traits<Graph>::vertex_descriptor>
 sprinkle (Graph const & g, unsigned int const n,
           unsigned int const max_adjacent, TimeoutTimer const & timeout)
{
  typedef boost::graph_traits<Graph> TR;
  typedef typename TR::vertex_descriptor V;

  std::vector<V> vs (vertices(g).first, vertices(g).second);

  for (;;)
  {
    timeout.check();

    std::random_shuffle(vs.begin(), vs.end());
    std::set<V> r;

    for (typename std::vector<V>::const_iterator j = vs.begin(); j != vs.end() && r.size() != n; ++j)
    {
      typename TR::adjacency_iterator i, e;
      for (boost::tie(i, e) = adjacent_vertices(*j, g); i != e; ++i)
      {
        // count already selected fields adjacent to this one:
        unsigned int c = 0;
        BOOST_FOREACH(Field const f, adjacent_vertices(*i, g))
          if (r.find(f) != r.end() && ++c == max_adjacent) goto too_crowded;
      }

      r.insert(*j);
      if (r.size() == n) return r;

      too_crowded:;
    }
  }
}

class MineObserver : public ObservedModel::ObserverBase<MineObserver, 0 /* order, sync: see notes.txt */>
{
  void init (ObservedModel const & model, TimeoutTimer const & t)
  {
    mined_property.resize(num_vertices(model.grid()));
    std::fill(mined_property.begin(), mined_property.end(), false);

    BOOST_FOREACH(Field const f, sprinkle(model.grid(), int(mineRatio * num_vertices(model.grid())), maxAdjacent, t))
      mined_property[f] = true;
  }

  MinedProperty mined_property;
  float mineRatio;
  unsigned int maxAdjacent;

  void onOperation (ObservedModel & model, Operation const & op)
  {
    if (RestartOperation const * const p = dynamic_cast<RestartOperation const *>(&op))
      init(model, p->t);
  }

  public:

    MineObserver (ObservedModel const & model, float const mr, unsigned int const ma, TimeoutTimer const & t)
      : mined_property (num_vertices(model.grid())),
        mineRatio (mr),
        maxAdjacent (ma)
    {
      init(model, t);
    }

    MineObserver (ObservedModel const & model, float const mr, unsigned int const ma, MinedProperty const & mp)
      : mined_property (mp),
        mineRatio (mr),
        maxAdjacent (ma)
    {
      assert(int(mineRatio * num_vertices(model.grid())) ==
              std::count(mined_property.begin(), mined_property.end(), true) &&
              "conflicting mined-property and mine-ratio passed");

      assert(mp.size() == num_vertices(model.grid()) && "conflicting model and mined-property passed");
    }

    friend MinedProperty const & M3D::mined (ObservedModel const &);
};

} // unnamed namespace

MinedProperty const & mined (ObservedModel const & model)
{
  return model.get<MineObserver>().mined_property;
}

void addMineObserver (ObservedModel & model, float const mineRatio,
                      unsigned int const maxAdjacentMines, TimeoutTimer const & t)
{
  model.add(MineObserver(model, mineRatio, maxAdjacentMines, t));
}

void addMineObserver (ObservedModel & model, float const mineRatio,
                      unsigned int const maxAdjacentMines, MinedProperty const & mp)
{
  model.add(MineObserver(model, mineRatio, maxAdjacentMines, mp));
}

unsigned int adjacentMines (ObservedModel const & model, Field const f)
{
  MinedProperty const & p = mined(model);
  unsigned int r = 0;
  BOOST_FOREACH(Field const g, adjacent_vertices(f, model.grid())) if (p[g]) ++r;
  return r;
}

} // namespace M3D
