
#ifndef M3D_OBSERVEDMODEL_HPP
#define M3D_OBSERVEDMODEL_HPP

#include <boost/multi_index/member.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index_container.hpp>

#include "BareModel.hpp"

#include "util/CloningPtr.hpp"
#include "util/TypeConstant.hpp"
#include "util/Clonable.hpp"
#include "util/VirtuallyDestructed.hpp"

namespace M3D {

struct ObservedModel : BareModel
{
  struct Observer : virtual Clonable<Observer>, VirtuallyDestructed
  /* Implementations:
  - Must *not* hold a reference to a model, since they may be copied to other Model instances.
  - Must not assume that changes they request will actually be made. Other observers or model classes in the hierarchy could override or revert them.
  - Must be deterministic in the sense that any particular model state should always trigger the same actions. If observers are not deterministic in this sense redo operations will not be reliable.
  */
  {
    virtual void onOperation (ObservedModel &, Operation const &) = 0;
      // Guaranteed to be called incrementally: for every change applied, all observers are notified before the next change is applied.
  };

  // Model constness propagates to observers.

  ObservedModel (BareModel const &);

  template <typename O> O const * getp () const { return static_cast<O const *>(getObserver(TypeToConstant<O>::value)); }
  template <typename O> O       * getp ()       { return static_cast<O       *>(getObserver(TypeToConstant<O>::value)); }

  template <typename O>
  O & get ()
  {
    O * const r = getp<O>();
    assert(r && "request for unregistered observed in observed model");
    return *r;
  }

  template <typename O>
  O const & get () const
  {
    O const * const r = getp<O>();
    assert(r && "request for unregistered observed in observed model");
    return *r;
  }

  template <typename Derived, int Order> struct ObserverBase : Observer, Clonable<Observer, Derived>
  {
    static int const order = Order;
  };

  template <typename O>
  void add (O const & o) // adds a copy of o
  {
    StoredObserver const t = { TypeToConstant<O>::value, O::order, CloningPtr<Observer>(new O (o)) };
    observers.insert(t);
  }

  template <typename O>
  void remove () { remove(TypeToConstant<O>::value); }

  virtual void apply (Operation const &);

  private:

    struct StoredObserver
    {
      TypeConstant cookie;
      int order;
      CloningPtr<Observer> instance;
    };

    struct cookie;
    struct order;

    typedef boost::multi_index_container<
      StoredObserver,
      boost::multi_index::indexed_by<
        boost::multi_index::ordered_unique<boost::multi_index::tag<cookie>,
                                           BOOST_MULTI_INDEX_MEMBER(StoredObserver, TypeConstant, cookie)>,
          // Deliberately unique; add(someObserver) must fail (and have no effect) if an observer of the given type is already present.
        boost::multi_index::ordered_non_unique<boost::multi_index::tag<order>,
                                               BOOST_MULTI_INDEX_MEMBER(StoredObserver, int, order)>
       >
     > Observers;

    // Ideally we should also have StoredConstObserver and ConstObservers, but we don't really need those at the moment so they're not implemented.

    bool applying;
    Observers observers;
    std::list<CloningPtr<Operation> > pending;

    void remove (TypeConstant);

    Observer * getObserver (TypeConstant) const;
};

template <typename F, int Order>
class ForwardingObserver : public ObservedModel::ObserverBase<ForwardingObserver<F, Order>, Order>
{
  F f;

  void onOperation (ObservedModel & m, Operation const & o) { f(m, o); }

  public:

    ForwardingObserver (F const f_) : f (f_) {}
};

template <int I, typename F>
ForwardingObserver<F, I> forwardingObserver (F const f) { return ForwardingObserver<F, I>(f); }

} // namespace M3D

#endif // header guard
