
#ifndef M3D_OBSERVERS_HPP
#define M3D_OBSERVERS_HPP

#include <boost/variant/static_visitor.hpp>
#include <algorithm>
#include <vector>

#include "ObservedModel.hpp" // not directly needed by this header (a fwd decl would suffice), but all includers need it anyway.
#include "Operations.hpp"

struct TimeoutTimer;

namespace M3D {

/* These Observers:
  - do not register/unregister themselves upon construction/destruction, they must be registered by their owners.
  - can be registered after model actions have taken place (they don't have to be present "from the start").
  - are only suitable for observing a single model.
*/

// Visualizations

  enum FieldVisualization
  {
    empty_vis,
    digit1_vis,
    digit2_vis,
    digit3_vis,
    digit4_vis,
    digit5_vis,
    digit6_vis,
    digit7_vis,
    digit8_vis,
    digit9_vis,
    neutral_vis,
    flag_vis,
    potentialMine_vis,
    mine_vis,
    falseFlag_vis,
    fatalMine_vis
  };

  typedef std::vector<FieldVisualization> Visualizations;

  bool hasFieldVisualizationSelector (ObservedModel const &);
  Visualizations const & visualization (ObservedModel const &);
  void addFieldVisualizationSelector (ObservedModel &, bool subtractFlagged);
  void removeFieldVisualizationSelector (ObservedModel &);

// Opening

  typedef std::vector<bool> OpenedProperty;

  enum PlayState { playing, won, lost };

  PlayState playState (ObservedModel const &);

  void addOpenObserver (ObservedModel &);
  void addOpenObserver (ObservedModel &, OpenedProperty const &);

  OpenedProperty const & opened (ObservedModel const &);
  void open (Field, ObservedModel &); // guards invalid attempts. has no effect if attempt invalid

// Flagging

  typedef std::vector<FlagLevel> FlaggedProperty;

  class FlagObserver : public ObservedModel::ObserverBase<FlagObserver, 0 /* order, sync: see notes.txt */>
  {
    FlaggedProperty flaggedProperty;
    bool cycleThroughRisky;

    void onOperation (ObservedModel &, Operation const &);

    public:

      void setCycleThroughRisky (bool const b) { cycleThroughRisky = b; }

      friend FlaggedProperty const & M3D::flagLevel (ObservedModel const &);

      void cycleFlagLevel (Field, ObservedModel &);

      FlagObserver (ObservedModel const & model, bool const cycleThroughRisky_,
                           FlaggedProperty const * const fp = 0)
        : flaggedProperty (fp ? *fp : FlaggedProperty(num_vertices(model.grid()), none)),
          cycleThroughRisky (cycleThroughRisky_)
      {
        assert(!fp || num_vertices(model.grid()) == fp->size()); // check for conflicting model and flagged-property
      }
  };

  FlaggedProperty const & flagLevel (ObservedModel const &);

  void cycleFlagLevel (Field, ObservedModel &);
  void flag (Field, ObservedModel &);
    // both guard invalid attempts and have no effect if attempt invalid

// Mining

  typedef std::vector<bool> MinedProperty;

  MinedProperty const & mined (ObservedModel const &);
  unsigned int adjacentMines (ObservedModel const &, Field);

  void addMineObserver (ObservedModel &, float mineRatio, unsigned int maxAdjacentMines, TimeoutTimer const &);
  void addMineObserver (ObservedModel &, float mineRatio, unsigned int maxAdjacentMines, MinedProperty const &);

// Misc

  void restartUntilPlaying (ObservedModel &, TimeoutTimer const &);
   // Basic error safety: no resources will be leaked, but it is destructive: the model should no longer be used if timeout or other exceptions occur.

  void addMinimalAutoOpener (ObservedModel &);
  void addNormalAutoOpener (ObservedModel &);
  void addAutoFlagger (ObservedModel &);
  void addAutoStarter (ObservedModel &, TimeoutTimer const &);

  void removeMinimalAutoOpener (ObservedModel &);
  void removeNormalAutoOpener (ObservedModel &);
  void removeAutoFlagger (ObservedModel &);
  void removeAutoStarter (ObservedModel &);

} // namespace M3D

#endif // header guard
