
#include "Observers.hpp"
#include "Operations.hpp"

#include <algorithm> // fill

#include "util/foreach.hpp"

namespace M3D {

namespace {

// OpenObserver's PlayState caching assumes that if the mine distribution changes, a RestartOperation will be sent.

class OpenObserver : public ObservedModel::ObserverBase<OpenObserver, -10 /* order, sync: see notes.txt */>,
                     Visitor<OpenOperation const>,
                     Visitor<RestartOperation const>
{
  ObservedModel const * model;
  OpenedProperty opened_property;
  PlayState playState; // Just a cache for efficiency. Invariant: corresponds to opened_property

  void visit (OpenOperation const & o)
  {
    if (playState == playing || (opened_property[o.field] && !o.opened)) // latter alternative is needed so we can undo from won/lost
      opened_property[o.field] = o.opened;
    updatePlayState();
  }

  void visit (RestartOperation const &)
  {
    std::fill(opened_property.begin(), opened_property.end(), false);
    updatePlayState();
  }

  void onOperation (ObservedModel & model_, Operation const & op)
  {
    model = &model_; // we might have moved
    op.accept(*this);
  }

  void updatePlayState ()
  {
    playState = won;
    BOOST_FOREACH(Field const f, vertices(model->grid()))
      if (opened_property[f] && mined(*model)[f]) { playState = lost; return; }
      else if (!opened_property[f] && !(mined(*model)[f])) playState = playing;
  }

  public:

    explicit OpenObserver (ObservedModel const & m)
      : model (&m), opened_property (num_vertices(m.grid()), false), playState (playing)
    {}

    OpenObserver (ObservedModel const & m, OpenedProperty const & p)
      : model (&m), opened_property (p)
    {
      assert(num_vertices(m.grid()) == p.size());
      updatePlayState();
    }

    friend PlayState M3D::playState (ObservedModel const &);
    friend OpenedProperty const & M3D::opened (ObservedModel const &);
};

} // unnamed namespace

OpenedProperty const & opened (ObservedModel const & model)
{
  return model.get<OpenObserver>().opened_property;
}

PlayState playState (ObservedModel const & model)
{
  return model.get<OpenObserver>().playState;
}

void open (Field const f, ObservedModel & model)
{
  if (opened(model)[f] || playState(model) != playing) return;

  OpenOperation o;
  o.field = f;
  o.opened = true;
  model.apply(o);
}

void addOpenObserver (ObservedModel & model)
{
  model.add(OpenObserver(model));
}

void addOpenObserver (ObservedModel & model, OpenedProperty const & op)
{
  model.add(OpenObserver(model, op));
}

void restartUntilPlaying (ObservedModel & m, TimeoutTimer const & t)
{
  while (playState(m) != playing)
  {
    t.check();
    restart(m, t);
  }
}

} // namespace M3D
