
// Previous versions of M3D had the notion of a "nullifies" relation between operations. When an operation O was added to an ObservedModel's pending queue, all existing pending operations P for which O.nullifies(P) held were removed. Even older versions of M3D had the notion of a "conflictsWith" relation that basically did the same thing. While these two relations had some uses in getting rid of redundant operations that were not direct duplicates of eachother, these advantages were mostly lost when the operation types themselves were changed to reflect the intended result state rather than the recipe to get to that state (i.e. when FlagOperation stored flagLevel rather than cycle, redundant but unequal flag operations became mostly harmless). Since then, the design was changed to simply use equality to remove of duplicate operations, because it is simpler and suffices for our needs. Still, we did lose some nice optimizations such having a RestartOperation nullify all existing pending operations.

#ifndef M3D_OPERATIONS_HPP
#define M3D_OPERATIONS_HPP

#include "util/TimeoutTimer.hpp"
#include "util/Clonable.hpp"
#include "util/VirtuallyDestructed.hpp"
#include "util/EqualityComparable.hpp"
#include "util/visitation.hpp"

#include "Grid.hpp"

namespace M3D {

struct ObservedModel;

struct Operation : VirtuallyDestructed,
                   virtual Clonable<Operation>,
                   virtual Visitable<>,
                   virtual EqualityComparable<>

{
  virtual Operation * inverse (ObservedModel const &) const { return 0; } // by default, operations are irriversible
};

template <typename Derived>
struct OperationBase : EqualityComparable<Derived>,
                       Clonable<Operation, Derived>,
                       Visitable<Derived>,
                       virtual Operation {};

// actual operations:

  struct OpenOperation : OperationBase<OpenOperation>
  {
    Field field;
    bool opened;

    Operation * inverse (ObservedModel const &) const; // todo: make return type OpenOperation* (SUSE93's gcc 3.3.5 chokes on it)

    bool operator== (OpenOperation const & other) const
    {
      return field == other.field && opened == other.opened;
    }
  };

  struct RestartOperation : OperationBase<RestartOperation>
  {
    explicit RestartOperation (TimeoutTimer const & t_) : t (t_) {}

    TimeoutTimer t;

    bool operator== (RestartOperation const &) const { return true; }

    // By not implementing inverse() this becomes an irreversable operation.
  };

  enum FlagLevel { flagged, risky, none, flagLevel_count };

  struct FlagOperation : OperationBase<FlagOperation>
  {
    Field field;
    FlagLevel flagLevel;

    Operation * inverse (ObservedModel const &) const; // todo: make return type OpenOperation* (SUSE93's gcc 3.3.5 chokes on it)

    bool operator== (FlagOperation const & other) const
    {
      return field == other.field && flagLevel == other.flagLevel;
    }
  };

// operation submitters

  void restart (ObservedModel &, TimeoutTimer const &);

} // namespace M3D

#endif // header guard
