
#include "Rotators.hpp"
#include "Settings.hpp"

#include "util/gettext.hpp"
#include "util/serialization-symbols.hpp"

#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix_expression.hpp>

namespace M3D {

namespace {

typedef boost::numeric::ublas::c_vector<float, 3> V;

V make_v (float x, float y, float z) { V r; r(0) = x; r(1) = y; r(2) = z; return r; }

V cross_prod (V a, V b)
{
  return make_v(a(1) * b(2) - a(2) * b(1),
                -(a(2) * b(0) - a(0) * b(2)),
                a(0) * b(1) - a(1) * b(0));
}

float length (V const & v)
{
  return std::sqrt(v(0) * v(0) + v(1) * v(1) + v(2) * v(2));
}

V normalize (V const & v) { return v / length(v); }


template <typename float_t>
boost::math::quaternion<float_t> rotation_quaternion (float_t const angle, V axis)
{
  float_t const sa = std::sin(angle / 2);
  axis *= sa;
  return boost::math::quaternion<float_t>(std::cos(angle / 2), axis(0), axis(1), axis(2));
}

float const default_y_axis_rotation = M_PI / 4,
            default_x_axis_rotation = -M_PI / 8;
  // Rotation sequencing is not commutative. Y axis rotation must be performed first.

boost::math::quaternion<float> default_rotation ()
{
  return rotation_quaternion<float>(default_y_axis_rotation, make_v(0, 1, 0)) *
         rotation_quaternion<float>(default_x_axis_rotation, make_v(1, 0, 0));
}

class NormalRotator : public Rotator
{
  float speed;

  boost::math::quaternion<float> q;

  void onDrag (float const xfrom, float const xto, float const yfrom, float const yto)
  {
    float const distance = std::sqrt((xto - xfrom) * (xto - xfrom) + (yto - yfrom) * (yto - yfrom));

    q *= rotation_quaternion<float>(-distance * speed, make_v((yto - yfrom) / distance,
                                                              (xto - xfrom) / distance, 0));
  }

  boost::math::quaternion<float> getRotation () const { return q; }

  public:

    explicit NormalRotator (float const s, Rotator const * const prev)
      : speed (s), q (prev ? prev->getRotation() : default_rotation())
    {}
};

class RestrictedRotator : public Rotator
{
  float y_axis_rotation, x_axis_rotation, speed;

  boost::math::quaternion<float> getRotation () const
  {
    return rotation_quaternion<float>(y_axis_rotation, make_v(0, 1, 0)) *
           rotation_quaternion<float>(x_axis_rotation, make_v(1, 0, 0));
  }

  void onDrag (float const xfrom, float const xto, float const yfrom, float const yto)
  {
    y_axis_rotation -= (xto - xfrom) * speed;

    x_axis_rotation = std::min(
                      std::max(
                        x_axis_rotation - (yto - yfrom) * speed,
                       -max_x_axis_rotation),
                       max_x_axis_rotation);
  }

  public:

    explicit RestrictedRotator (float const s, Rotator const * const prev)
      : y_axis_rotation (default_y_axis_rotation),
        x_axis_rotation (default_x_axis_rotation),
        speed (s)
    {
      if (prev)
        if (RestrictedRotator const * const p = dynamic_cast<RestrictedRotator const *>(prev))
        {
          y_axis_rotation = p->y_axis_rotation;
          x_axis_rotation = p->x_axis_rotation;
        }
    }

    static float const max_x_axis_rotation;
};

float const RestrictedRotator::max_x_axis_rotation = M_PI / 2;

class TrackballRotator : public Rotator
{
  // This rotator simulates dragging a big sphere the size of the viewing area. The sphere doesn't cover the complete viewing area, because, well, spheres aren't rectangular. This raises the question of what to do when the user drags outside the circle. In this implementation, this is handled by pretending that the user dragged at the nearest edge of the circle. This rotator isn't speed-configurable, because there is a direct mapping from drag start and end points to points on the virtual sphere, and the amount dragged is exactly the difference between those points.

  static V sphere_point (float x, float y) // assumes a sphere with radius 1
  {
    float const d = sqrt(x * x + y * y), distance_cutoff = 0.9;
      // Obviously ideally distance_cutoff would just be 1, but we need a bit of margin to make sure sqrt will not be called on a negative value due to the limited-precision nature of floats.
    if (d > distance_cutoff) { x = x / d * distance_cutoff; y = y / d * distance_cutoff; }
    return make_v(x, y, std::sqrt(1 - x * x - y * y));
  }

  boost::math::quaternion<float> q;

  void onDrag (float const xfrom, float const xto, float const yfrom, float const yto)
  {
    V const from = sphere_point(xfrom * 2, yfrom * 2),
            to = sphere_point(xto * 2, yto * 2);

    // from and to are already unit vectors, so we don't have to divide the inner product by them before acos-ing to get the angle

    q *= rotation_quaternion(std::acos(inner_prod(from, to)),
                             normalize(cross_prod(from, to)));
  }

  boost::math::quaternion<float> getRotation () const { return q; }

  public:

    explicit TrackballRotator (float const /* speed */, Rotator const * const prev)
      : q (prev ? prev->getRotation() : default_rotation())
    {}
};

template <typename RotatorT>
struct RotatorFactoryImpl : RotatorFactory
{
  RotatorT * create (float const speed, Rotator const * const prev) const { return new RotatorT (speed, prev); }
};

template <typename T> T & static_instance_of () { static T r; return r; }

} // unnamed namespace

serialization_symbols<RotatorFactory const *>::type make_serialization_symbols (
  SerializationSymbolOverloadTag<RotatorFactory const *, CanvasSettings, &CanvasSettings::rotation>)
{
  serialization_symbols<RotatorFactory const *>::type r;
  r.insert(std::make_pair(&static_instance_of<RotatorFactoryImpl<RestrictedRotator> >(), N_("restricted")));
  r.insert(std::make_pair(&static_instance_of<RotatorFactoryImpl<NormalRotator    > >(), N_("normal")));
  r.insert(std::make_pair(&static_instance_of<RotatorFactoryImpl<TrackballRotator > >(), N_("trackball")));
  return r;
}

} // namespace M3D
