
// While dialog() could simply depend on validate(*::Settings) to do validation, it's more user friendly to make the widgets behave in such a way that entering an invalid configuration is simply not possible. This is achieved by automatically modifying other settings when the user changes certain settings that would otherwise conflict.

#include <stdexcept> // runtime_error
#include <boost/archive/xml_iarchive.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/noncopyable.hpp>
#include <boost/none.hpp>
#include <boost/bind.hpp>
#include <boost/serialization/nvp.hpp>
#include "application-constants.hpp"
#include "Application.hpp"
#include "util/foreach.hpp"
#include "util/ScopeGuard.hpp"
#include "util/glademm.hpp"
#include "util/bfstream.hpp"
#include "util/gettext.hpp"

#include "Settings.hpp"
  // We'd like this to be the first include, but Boost.Serialization has some strange header order requirements that won't allow that.

namespace fs = boost::filesystem;

namespace M3D {

namespace {

template <typename T, typename Pod, T (Pod:: * M), bool apply_gettext_to_symbols = true>
class SerializedComboBox : public Gtk::ComboBox
// A combo box that displays items from a serialization_symbols map. The (optionally gettext'ed) symbol is used for display while the values are stored underneath.
{
  struct Columns : Gtk::TreeModel::ColumnRecord
  {
    Columns () { add(name); add(value); }
    Gtk::TreeModelColumn<std::string> name;
    Gtk::TreeModelColumn<T> value;
  } const columns;

  public:

    SerializedComboBox (BaseObjectType * const b, Glib::RefPtr<Gnome::Glade::Xml>)
      : Gtk::ComboBox (b)
    {
      Glib::RefPtr<Gtk::ListStore> const list = Gtk::ListStore::create(columns);
      typedef std::pair<T, std::string> P;
      typedef typename serialization_symbols<T>::type SS;
      SS const & ss = get_serialization_symbols<T, Pod, M>();

      for (typename SS::const_iterator i = ss.begin(); i != ss.end(); ++i)
      {
        Gtk::TreeModel::Row const r = *(list->append());
        r[columns.value] = i->first;
        r[columns.name] = (apply_gettext_to_symbols ? gettext(i->second.c_str()) : i->second);
      }

      set_model(list);
      pack_start(columns.name);
    }

    void select (T const v)
    {
      Glib::RefPtr<Gtk::TreeModel> const m = get_model();
      Gtk::TreeModel::Children const c = m->children();
      for (Gtk::TreeModel::iterator i = c.begin(); i != c.end(); ++i)
        if (T((*i)[columns.value]) == v) { set_active(i); return; }
      throw std::runtime_error(_("Invalid value selected."));
    }

    T get () const { return (*get_active())[columns.value]; }
};

struct ScrolledWindow : Gtk::ScrolledWindow
{
  // Fragile: I wrote this with very limited knowledge of the Gtk size negotion system.

  void on_size_request (Gtk::Requisition * const r)
  {
    // The problem with the default on_size_request is that it throws the child's requested size out the window unless the scrollbar policy is "never". Ideally we'd like to just use the child's requested size plus some extra space for the scrollbars, but for some reason if we don't actually call the inherited on_size_request the scrollbars aren't shown at all.

    GtkRequisition t;
    Gtk::ScrolledWindow::on_size_request(&t);
    gtk_widget_size_request (get_child()->gobj(), r);
    gtk_widget_size_request (boost::implicit_cast<Gtk::Widget *>(get_hscrollbar())->gobj(), &t);
    r->height += t.height;
    gtk_widget_size_request (boost::implicit_cast<Gtk::Widget *>(get_vscrollbar())->gobj(), &t);
    r->width += t.width;
  }

  ScrolledWindow (BaseObjectType * const b, Glib::RefPtr<Gnome::Glade::Xml> const &) : Gtk::ScrolledWindow (b) {}
};

class PreferencesDialog : public Gtk::Dialog
{
  Glib::RefPtr<Gnome::Glade::Xml> gladeXml;

  template <typename T> T & get (std::string const & s) const
  {
    using ::get; // workaround for gcc 3.3.5 name lookup bug
    return ::get<T>(gladeXml, s);
  }

  void on_subtractFlagged_toggle ()
  {
    if (get<Gtk::CheckButton>("subtractFlagged").get_active())
    {
      get<Gtk::CheckButton>("autoOpen").set_active(true);
      on_autoOpen_toggle();
    }
  }

  void on_potentialMineMarkers_toggle ()
  {
    if (get<Gtk::CheckButton>("potentialMineMarkers").get_active())
    {
      get<Gtk::CheckButton>("autoOpen").set_active(false);
      on_autoOpen_toggle();
    }
  }

  void on_autoOpen_toggle ()
  {
    if (get<Gtk::CheckButton>("autoOpen").get_active())
    {
      get<Gtk::CheckButton>("potentialMineMarkers").set_active(false);
      on_potentialMineMarkers_toggle();
    }
    else
    {
      get<Gtk::CheckButton>("subtractFlagged").set_active(false);
      on_subtractFlagged_toggle();
    }
  }

  typedef SerializedComboBox<GridShape const *, GridSettings, &GridSettings::gridShape> GridShape_box;
  typedef SerializedComboBox<FieldShape const *, GridSettings, &GridSettings::fieldShape> FieldShape_box;
  typedef SerializedComboBox<RotatorFactory const *, CanvasSettings, &CanvasSettings::rotation> Rotation_box;
  typedef SerializedComboBox<HighlightStyle, CanvasSettings, &CanvasSettings::highlightStyle> Highlighting_box;
  typedef SerializedComboBox<fs::path, CanvasSettings, &CanvasSettings::skin, false /* no gettext */> Skin_box;

  GridShape_box & gridShape;
  FieldShape_box & fieldShape;
  Rotation_box & rotation;
  Highlighting_box & highlighting;
  Skin_box & skin;
  ScrolledWindow & scrolledWindow;

  public:

    PreferencesDialog (BaseObjectType * const b,
                        Glib::RefPtr<Gnome::Glade::Xml> const & gladeXml_)
      : Gtk::Dialog (b), gladeXml (gladeXml_),
        gridShape      (get_derived<GridShape_box   >(gladeXml, "gridShape"     )),
        fieldShape     (get_derived<FieldShape_box  >(gladeXml, "fieldShape"    )),
        rotation       (get_derived<Rotation_box    >(gladeXml, "rotation"      )),
        highlighting   (get_derived<Highlighting_box>(gladeXml, "highlighting"  )),
        skin           (get_derived<Skin_box        >(gladeXml, "skin"          )),
        scrolledWindow (get_derived<ScrolledWindow  >(gladeXml, "scrolledWindow"))
    {
      get<Gtk::CheckButton>("autoOpen").signal_toggled().
        connect(boost::bind(&PreferencesDialog::on_autoOpen_toggle, this));
      get<Gtk::CheckButton>("subtractFlagged").signal_toggled().
        connect(boost::bind(&PreferencesDialog::on_subtractFlagged_toggle, this));
      get<Gtk::CheckButton>("potentialMineMarkers").signal_toggled().
        connect(boost::bind(&PreferencesDialog::on_potentialMineMarkers_toggle, this));
    }

    void set (Settings const & s)
    {
      rotation.select(s->canvasSettings.rotation);
      gridShape.select(s->modelSettings.gridSettings.gridShape);
      fieldShape.select(s->modelSettings.gridSettings.fieldShape);
      highlighting.select(s->canvasSettings.highlightStyle);
      skin.select(s->canvasSettings.skin);
      get<Gtk::CheckButton>("autoOpen").set_active(s->modelSettings.autoOpen);
      get<Gtk::CheckButton>("autoFlag").set_active(s->modelSettings.autoFlag);
      get<Gtk::CheckButton>("subtractFlagged").set_active(s->subtractFlagged);
      get<Gtk::CheckButton>("potentialMineMarkers").set_active(s->cycleThroughRisky);
      get<Gtk::SpinButton>( "gridSize").set_value(s->modelSettings.gridSettings.size);
      get<Gtk::SpinButton>("mineRatio").set_value(s->modelSettings.mineRatio);
      get<Gtk::CheckButton>("lessDetection").set_active(s->modelSettings.gridSettings.lessDetection);
      get<Gtk::HScale>("rotationSpeed").set_value(s->canvasSettings.rotationSpeed);
      get<Gtk::HScale>("fieldSize").set_value(s->canvasSettings.solidFieldSize);
      get<Gtk::HScale>("digitSize").set_value(s->canvasSettings.digitSize);
    }

    Settings getSettings (Settings::base_type r)
      // We start with a given existing settings object because there may be fields that we don't modify, but which must not be lost.
    {
      r.modelSettings.gridSettings.gridShape = gridShape.get();
      r.modelSettings.gridSettings.fieldShape = fieldShape.get();
      r.modelSettings.autoOpen = get<Gtk::CheckButton>("autoOpen").get_active();
      r.modelSettings.autoFlag = get<Gtk::CheckButton>("autoFlag").get_active();
      r.canvasSettings.highlightStyle = highlighting.get();
      r.canvasSettings.rotation = rotation.get();
      r.subtractFlagged = get<Gtk::CheckButton>("subtractFlagged").get_active();
      r.cycleThroughRisky = get<Gtk::CheckButton>("potentialMineMarkers").get_active();
      r.modelSettings.gridSettings.size = int(get<Gtk::SpinButton>("gridSize").get_value());
      r.modelSettings.mineRatio = get<Gtk::SpinButton>("mineRatio").get_value();
      r.modelSettings.gridSettings.lessDetection = get<Gtk::CheckButton>("lessDetection").get_active();
      r.canvasSettings.solidFieldSize = get<Gtk::HScale>("fieldSize").get_value();
      r.canvasSettings.rotationSpeed = get<Gtk::HScale>("rotationSpeed").get_value();
      r.canvasSettings.digitSize = get<Gtk::HScale>("digitSize").get_value();
      r.canvasSettings.skin = skin.get();

      return r;
    }
};

} // unnamed namespace

boost::optional<Settings> dialog (
  Settings const & settings, Glib::RefPtr<Gnome::Glade::Xml> const & gladeXml)
{
  PreferencesDialog & d = get_derived<PreferencesDialog>(gladeXml, "preferencesDialog");
  d.set(settings);
  boost::optional<Settings> r;
  d.show_all();
  ScopeGuard const hider (boost::bind(&PreferencesDialog::hide_all, &d));
  if (d.run() == Gtk::RESPONSE_OK) r = d.getSettings(*settings);
  return r;
}

void validate (Settings::base_type const & settings)
{
  if (settings.modelSettings.maxAdjacentMines > 9)
    throw ValidationError(_("Having a maximum number of adjacent mines higher than 9 is (currently) not supported."));

  if (settings.subtractFlagged &&
      !settings.modelSettings.autoOpen)
    throw ValidationError(_("The 'Subtract flag counts from mine counts' option can only be enabled if the 'Automatically open fields that are obviously clear' option is also enabled."));
}

Settings loadSettingsFrom (boost::filesystem::path const & p)
{
  boost::filesystem::ifstream f (p);
   // We don't use bifstream here, because users might edit their settings file manually, thereby inserting platform newlines.
  boost::archive::xml_iarchive a (f);
  Settings::base_type r;
  a >> boost::serialization::make_nvp("settings", r);
  return Settings(r);
}

void saveToFile (boost::filesystem::path const & p, Settings const & s)
{
  bofstream f (p);
  boost::archive::xml_oarchive a (f);
  a << boost::serialization::make_nvp("settings", s.base());
}

} // namespace M3D

serialization_symbols<fs::path>::type make_serialization_symbols (
  SerializationSymbolOverloadTag<fs::path, M3D::CanvasSettings, &M3D::CanvasSettings::skin>)
{
  ::serialization_symbols<fs::path>::type r;

  fs::path const a (M3D::app_user_path() / "skins"),
                 b (APP_RESOURCE_PATH "/skins", fs::native);

  assert(exists(a));
  assert(exists(b));

  for (fs::directory_iterator i (a), e; i != e; ++i)
    if (!fs::is_directory(*i)) r.insert(std::make_pair(*i, i->leaf()));
  for (fs::directory_iterator i (b), e; i != e; ++i)
    if (!fs::is_directory(*i)) r.insert(std::make_pair(*i, i->leaf()));
    // Todo: Verify that this works (with proper overriding).

  return r;
}
