
#ifndef M3D_SETTINGS_HPP
#define M3D_SETTINGS_HPP

#include <string>
#include <boost/optional.hpp>
#include <boost/noncopyable.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/implicit_cast.hpp>

#include <libglademm.h>

#include "geometry.hpp"
#include "Rotators.hpp"

#include "util/color.hpp"
#include "util/gettext.hpp"
#include "util/Validated.hpp"
#include "util/pod-definition.hpp"
#include "util/pod-serialization.hpp"
#include "util/pod-equality.hpp"
#include "util/enum-definition.hpp"
#include "util/enum-serialization-symbols.hpp"

namespace M3D {

DEFINE_WRAPPED_ENUM(HighlightStyle, (( none     ))\
                                    (( single   ))\
                                    (( adjacent )))
  // Register for gettext:
  R_("none") R_("single") R_("adjacent")

ENUM_SERIALIZATION_SYMBOLS(HighlightStyle)

POD_DEFINE(GridSettings, (( FieldShape const *, fieldShape ))\
                         (( GridShape const *, gridShape   ))\
                         (( unsigned int, size             ))\
                         (( bool, lessDetection            )))

serialization_symbols<M3D::FieldShape const *>::type make_serialization_symbols
  (SerializationSymbolOverloadTag<M3D::FieldShape const *, GridSettings, &GridSettings::fieldShape>);

POD_DEFINE(ModelSettings, (( GridSettings, gridSettings     ))\
                          (( float, mineRatio               ))\
                          (( unsigned int, maxAdjacentMines ))\
                          (( bool, autoFlag                 ))\
                          (( bool, autoOpen                 )))

POD_DEFINE(CanvasSettings, (( HighlightStyle, highlightStyle    ))\
                           (( RotatorFactory const *, rotation  ))\
                           (( float, rotationSpeed              ))\
                           (( float, solidFieldSize             ))\
                           (( float, digitSize                  ))\
                           (( float, fieldOfVision              ))\
                           (( unsigned int, width               ))\
                           (( unsigned int, height              ))\
                           (( unsigned int, textureSize         ))\
                           (( boost::filesystem::path, skin     )))

serialization_symbols<RotatorFactory const *>::type make_serialization_symbols (
  SerializationSymbolOverloadTag<RotatorFactory const *, CanvasSettings, &CanvasSettings::rotation>);

POD_DEFINE(Settings_base, (( ModelSettings, modelSettings   ))\
                          (( CanvasSettings, canvasSettings ))\
                          (( bool, subtractFlagged          ))\
                          (( bool, cycleThroughRisky        )))

typedef Validated<Settings_base> Settings;

void validate (Settings::base_type const &);

Settings loadSettingsFrom (boost::filesystem::path const &);
void saveToFile (boost::filesystem::path const &, Settings const &);

boost::optional<Settings> dialog (Settings const &, Glib::RefPtr<Gnome::Glade::Xml> const &);
  // - shows a modal dialog that lets the user modify the settings
  // - returns new settings iff the user accepted the settings

POD_DEFINE_EQUALITY(GridSettings)
POD_DEFINE_EQUALITY(ModelSettings)

} // namespace M3D

// These two outside M3D because their origins lie elsewhere and ADL won't be able to find them in M3D.

serialization_symbols<boost::filesystem::path>::type make_serialization_symbols (
  SerializationSymbolOverloadTag<boost::filesystem::path, M3D::CanvasSettings, &M3D::CanvasSettings::skin>);

serialization_symbols<M3D::GridShape const *>::type make_serialization_symbols
 (SerializationSymbolOverloadTag<M3D::GridShape const *, M3D::GridSettings, &M3D::GridSettings::gridShape>);

POD_DEFINE_SERIALIZATION(M3D::GridSettings)
POD_DEFINE_SERIALIZATION(M3D::ModelSettings)
POD_DEFINE_SERIALIZATION(M3D::CanvasSettings)
POD_DEFINE_SERIALIZATION(M3D::Settings::base_type)

BOOST_CLASS_IMPLEMENTATION(M3D::GridSettings, boost::serialization::object_serializable) \
BOOST_CLASS_IMPLEMENTATION(M3D::ModelSettings, boost::serialization::object_serializable) \
BOOST_CLASS_IMPLEMENTATION(M3D::CanvasSettings, boost::serialization::object_serializable) \
BOOST_CLASS_IMPLEMENTATION(M3D::Settings::base_type, boost::serialization::object_serializable) \

BOOST_CLASS_TRACKING(M3D::GridSettings, boost::serialization::track_never)
BOOST_CLASS_TRACKING(M3D::ModelSettings, boost::serialization::track_never)
BOOST_CLASS_TRACKING(M3D::CanvasSettings, boost::serialization::track_never)
BOOST_CLASS_TRACKING(M3D::Settings::base_type, boost::serialization::track_never)

#endif // header guard
