
#include "geometry.hpp"
#include "Settings.hpp"
#include <algorithm> // max
#include <cassert>
#include <boost/assign/list_of.hpp>
#include "util/gettext.hpp"

namespace M3D {

FieldShape::Side::Side (Vertex const n, Vertex const c, Vertex const q0,
                        Vertex const q1, Vertex const q2, Vertex const q3)
  : normal (n), center (c), quad (boost::assign::list_of(q0)(q1)(q2)(q3))
{}

namespace {

template <typename T>
boost::numeric::ublas::c_vector<T, 3> cvec3 (T const a, T const b, T const c)
{
  boost::numeric::ublas::c_vector<T, 3> r; r(0) = a; r(1) = b; r(2) = c; return r;
}

bool sphereCarvePredicate (FieldCoordinates const & c, unsigned int const size)
{
  return inner_prod(c, c) < int(size * size);
}

bool cubeCarvePredicate (FieldCoordinates const &, unsigned int const size) { return true; }

GridShape const sphereGridShape (sphereCarvePredicate),
                cubeGridShape (cubeCarvePredicate);

class CubeFieldShape : public FieldShape
{
  float spriteSize () const { return 2; }

  std::vector<FieldCoordinates> generateFieldCoordinates (int const size) const
  {
    std::vector<FieldCoordinates> r;
    FieldCoordinates t;
    for (t(0) = 1 - int(size); t(0) < size; ++++t(0))
    for (t(1) = 1 - int(size); t(1) < size; ++++t(1))
    for (t(2) = 1 - int(size); t(2) < size; ++++t(2))
      r.push_back(t);
    return r;
  }

  bool adjacent (FieldCoordinates const & a, FieldCoordinates const & b, bool const lessDetection) const
  {
    return (abs(a(0) - b(0)) <= 2 &&
            abs(a(1) - b(1)) <= 2 &&
            abs(a(2) - b(2)) <= 2 &&
            (!lessDetection || ((a(0) == b(0)) ||
                                (a(1) == b(1)) ||
                                (a(2) == b(2)))
           ));
  }

  public:

    CubeFieldShape ()
    {
      sides_.push_back(Side(cvec3( 0,  0, -1), cvec3( 0,  0, -1), cvec3(-1, -1, -1), cvec3(-1,  1, -1), cvec3( 1,  1, -1), cvec3( 1, -1, -1)));
      sides_.push_back(Side(cvec3( 0,  0,  1), cvec3( 0,  0,  1), cvec3(-1, -1,  1), cvec3( 1, -1,  1), cvec3( 1,  1,  1), cvec3(-1,  1,  1)));
      sides_.push_back(Side(cvec3( 0, -1,  0), cvec3( 0, -1,  0), cvec3(-1, -1, -1), cvec3( 1, -1, -1), cvec3( 1, -1,  1), cvec3(-1, -1,  1)));
      sides_.push_back(Side(cvec3( 0,  1,  0), cvec3( 0,  1,  0), cvec3(-1,  1, -1), cvec3(-1,  1,  1), cvec3( 1,  1,  1), cvec3( 1,  1, -1)));
      sides_.push_back(Side(cvec3(-1,  0,  0), cvec3(-1,  0,  0), cvec3(-1, -1, -1), cvec3(-1, -1,  1), cvec3(-1,  1,  1), cvec3(-1,  1, -1)));
      sides_.push_back(Side(cvec3( 1,  0,  0), cvec3( 1,  0,  0), cvec3( 1, -1, -1), cvec3( 1,  1, -1), cvec3( 1,  1,  1), cvec3( 1, -1,  1)));
    }
}
const cubeFieldShape;

class DodecahedronFieldShape : public FieldShape
{
  float spriteSize () const { return 1.5; }

  std::vector<FieldCoordinates> generateFieldCoordinates (int const size) const
  {
    std::vector<FieldCoordinates> r;
    FieldCoordinates t;

    for (t(0) = 1 - int(size); t(0) < size; ++++t(0))
    for (t(1) = 1 - int(size); t(1) < size; ++++t(1))
    for (t(2) = 1 - int(size); t(2) < size; ++++t(2))
    {
      r.push_back(t);
      if (t(0) + 2 < size && t(1) + 2 < size) r.push_back(cvec3(t(0) + 1, t(1) + 1, t(2)    ));
      if (t(0) + 2 < size && t(2) + 2 < size) r.push_back(cvec3(t(0) + 1, t(1)    , t(2) + 1));
      if (t(1) + 2 < size && t(2) + 2 < size) r.push_back(cvec3(t(0),     t(1) + 1, t(2) + 1));
    }
    return r;
  }

  bool adjacent (FieldCoordinates const & a, FieldCoordinates const & b, bool const lessDetection) const
  {
    if (abs(a(0) - b(0)) > 2 ||
        abs(a(1) - b(1)) > 2 ||
        abs(a(2) - b(2)) > 2) return false; // merely an optimization

    return (((a(0) == b(0) && abs(a(1) - b(1)) + abs(a(2) - b(2)) == 2) ||
             (a(1) == b(1) && abs(a(2) - b(2)) + abs(a(0) - b(0)) == 2) ||
             (a(2) == b(2) && abs(a(0) - b(0)) + abs(a(1) - b(1)) == 2))
            &&
            (!lessDetection || ((a(0) != b(0) && a(1) != b(1)) ||
                                (a(0) != b(0) && a(2) != b(2)) ||
                                (a(1) != b(1) && a(2) != b(2)))
           ));
  }

  public:

    DodecahedronFieldShape ()
    {
      sides_ = boost::assign::list_of
        (Side(cvec3( 0,  1,  1), cvec3( 0.0,  0.5,  0.5), cvec3(0,  1,  0), cvec3(-0.5,  0.5,  0.5), cvec3( 0, 0,  1), cvec3( 0.5,  0.5,  0.5)))
        (Side(cvec3( 0,  1, -1), cvec3( 0.0,  0.5, -0.5), cvec3(0,  1,  0), cvec3( 0.5,  0.5, -0.5), cvec3( 0, 0, -1), cvec3(-0.5,  0.5, -0.5)))
        (Side(cvec3( 0, -1,  1), cvec3( 0.0, -0.5,  0.5), cvec3(0, -1,  0), cvec3( 0.5, -0.5,  0.5), cvec3( 0, 0,  1), cvec3(-0.5, -0.5,  0.5)))
        (Side(cvec3( 0, -1, -1), cvec3( 0.0, -0.5, -0.5), cvec3(0, -1,  0), cvec3(-0.5, -0.5, -0.5), cvec3( 0, 0, -1), cvec3( 0.5, -0.5, -0.5)))
        (Side(cvec3( 1,  0,  1), cvec3( 0.5,  0.0,  0.5), cvec3(0,  0,  1), cvec3( 0.5, -0.5,  0.5), cvec3( 1, 0,  0), cvec3( 0.5,  0.5,  0.5)))
        (Side(cvec3(-1,  0,  1), cvec3(-0.5,  0.0,  0.5), cvec3(0,  0,  1), cvec3(-0.5,  0.5,  0.5), cvec3(-1, 0,  0), cvec3(-0.5, -0.5,  0.5)))
        (Side(cvec3( 1,  0, -1), cvec3( 0.5,  0.0, -0.5), cvec3(0,  0, -1), cvec3( 0.5,  0.5, -0.5), cvec3( 1, 0,  0), cvec3( 0.5, -0.5, -0.5)))
        (Side(cvec3(-1,  0, -1), cvec3(-0.5,  0.0, -0.5), cvec3(0,  0, -1), cvec3(-0.5, -0.5, -0.5), cvec3(-1, 0,  0), cvec3(-0.5,  0.5, -0.5)))
        (Side(cvec3( 1,  1,  0), cvec3( 0.5,  0.5,  0.0), cvec3(0,  1,  0), cvec3( 0.5,  0.5,  0.5), cvec3( 1, 0,  0), cvec3( 0.5,  0.5, -0.5)))
        (Side(cvec3(-1,  1,  0), cvec3(-0.5,  0.5,  0.0), cvec3(0,  1,  0), cvec3(-0.5,  0.5, -0.5), cvec3(-1, 0,  0), cvec3(-0.5,  0.5,  0.5)))
        (Side(cvec3( 1, -1,  0), cvec3( 0.5, -0.5,  0.0), cvec3(0, -1,  0), cvec3( 0.5, -0.5, -0.5), cvec3( 1, 0,  0), cvec3( 0.5, -0.5,  0.5)))
        (Side(cvec3(-1, -1,  0), cvec3(-0.5, -0.5,  0.0), cvec3(0, -1,  0), cvec3(-0.5, -0.5,  0.5), cvec3(-1, 0,  0), cvec3(-0.5, -0.5, -0.5)));
    }
}
const dodecahedronFieldShape;

} // unnamed namespace

serialization_symbols<M3D::FieldShape const *>::type make_serialization_symbols
  (SerializationSymbolOverloadTag<M3D::FieldShape const *, GridSettings, &GridSettings::fieldShape>)
{
  ::serialization_symbols<M3D::FieldShape const *>::type r;
  r.insert(std::make_pair(&M3D::dodecahedronFieldShape, N_("dodecahedron")));
  r.insert(std::make_pair(&M3D::cubeFieldShape, N_("cube")));
  return r;
}

} // namespace M3D

serialization_symbols<M3D::GridShape const *>::type make_serialization_symbols
 (SerializationSymbolOverloadTag<M3D::GridShape const *, M3D::GridSettings, &M3D::GridSettings::gridShape>)
{
  ::serialization_symbols<M3D::GridShape const *>::type r;
  r.insert(std::make_pair(&M3D::sphereGridShape, N_("sphere")));
  r.insert(std::make_pair(&M3D::cubeGridShape, N_("cube")));
  return r;
}
