
#ifndef M3D_GEOMETRY_HPP
#define M3D_GEOMETRY_HPP

#include <boost/function.hpp>
#include <boost/array.hpp>
#include <boost/static_assert.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "util/serialization-symbols.hpp"
#include "util/VirtuallyDestructed.hpp"

#include <vector>

namespace M3D {

typedef boost::numeric::ublas::c_vector<int, 3> FieldCoordinates; // relative to the center of the model

typedef boost::function<bool(FieldCoordinates const &, unsigned int size)> GridShape;
  // The essense of a grid shape is the logic that decides whether a given field is "carved" out of a cubic block of fields of a given size.

struct FieldShape : boost::noncopyable, VirtuallyDestructed
{
  typedef boost::numeric::ublas::c_vector<float, 3> Vertex;
  typedef boost::array<Vertex, 4> Quad; // currently, sides are always quads

  struct Side
  {
    Vertex normal, center;
    Quad quad;

    Side (Vertex n, Vertex c, Vertex q0, Vertex q1, Vertex q2, Vertex q3);
  };

  typedef std::vector<Side> Sides;

  Sides const & sides () const { return sides_; }

  virtual std::vector<FieldCoordinates> generateFieldCoordinates (int size) const = 0;
    // Generates a fully tesselated cube shaped set of field coordinates.
    // Todo: Accepting an output iterator seems more civilized than returning a std::vector, but we'd need something like any_iterator (wasn't that proposed for Boost already?)
  virtual bool adjacent (FieldCoordinates const &, FieldCoordinates const &, bool lessDetection) const = 0;
  virtual float spriteSize () const = 0;

  protected:

    Sides sides_;
};

// FieldShape's are immutable and only some static duration instances are used, which makes mapping serialization symbols really easy since we can simply immediately map the pointers to those instances.

} // namespace M3D

#endif // header guard
