
#ifndef UTIL_FIXEDVECTOR_HPP
#define UTIL_FIXEDVECTOR_HPP

#include <boost/array.hpp>
#include <algorithm> // equal

template <typename T, std::size_t MaxSize>
class FixedVector : boost::array<T, MaxSize>
/* - has std::vector's interface
   - has a template parameter which specifies the maximum sequence size
   - stores elements in the FixedVector object itself, no heap/free-store allocation is performed. This means that sizeof(FixedVector<T, N>) will be at least sizeof(T)*N.
*/
{
  // FixedVector can theoretically support all of std::vector's operations. However, only the ones actually used by M3D are currently implemented.

  // M3D only uses FixedVector for simple integral vertex descriptors. For these, it is not a problem that the current FixedVector implementation default constructs all sequence elements at initialization and doesn't destruct removed elements. However, for more complex types these would be problems.

  typedef boost::array<T, MaxSize> Base;

  public:

    #define USE_BASE_TYPE(t) typedef typename Base::t t;

    USE_BASE_TYPE(value_type)
    USE_BASE_TYPE(reference)
    USE_BASE_TYPE(const_reference)
    USE_BASE_TYPE(iterator)
    USE_BASE_TYPE(const_iterator)
    USE_BASE_TYPE(difference_type)
    USE_BASE_TYPE(size_type)
    USE_BASE_TYPE(reverse_iterator)
    USE_BASE_TYPE(const_reverse_iterator)

    #undef USE_BASE_TYPE

    using Base::operator[];
    using Base::begin;
    using Base::rend;
    using Base::front;

  private:

    size_type size_;
    using Base::elems;

  public:

    template <typename I> FixedVector (I i, I const e) : size_ (0) { for (; i != e; ++i) push(*i); }

    FixedVector () : size_ (0) {}

    explicit FixedVector (size_type const s) : size_ (s) {}

          iterator end ()       { return begin() + size_; }
    const_iterator end () const { return begin() + size_; }

          reverse_iterator rbegin ()       { return       reverse_iterator(end()); }
    const_reverse_iterator rbegin () const { return const_reverse_iterator(end()); }

          reference at (size_type const i)       { if (i < size_) return elems[i]; else throw std::out_of_range(); }
    const_reference at (size_type const i) const { if (i < size_) return elems[i]; else throw std::out_of_range(); }

          reference back ()       { assert(size_ > 0); return elems[size_ - 1]; }
    const_reference back () const { assert(size_ > 0); return elems[size_ - 1]; }

    size_type capacity () const { return MaxSize; }
    size_type max_size () const { return MaxSize; }

    void push (T const & t) { assert(size_ != MaxSize); elems[size_] = t; ++size_; }
    void pop () { assert(size_ != 0); --size_; }
    void push_back (T const & t) { push(t); }
    void pop_back () { pop(); }

    size_type size () const { return size_; }

    bool empty () const { return size_ == 0; }
};

template <typename T, typename U, std::size_t MaxSize>
bool operator== (FixedVector<T, MaxSize> const & a, FixedVector<U, MaxSize> const & b)
{
  return a.size() == b.size() && std::equal(a.begin(), a.end(), b.begin());
}

#endif // header guard
