
#ifndef GLUTILS_HPP
#define GLUTILS_HPP

#include <GL/gl.h>

#include <boost/noncopyable.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <algorithm>

typedef boost::numeric::ublas::c_vector<GLfloat, 4> GLvector;
typedef boost::numeric::ublas::c_matrix<GLfloat, 4, 4> GLmatrix;

inline GLenum glGetMatrixMode () { GLint r; glGetIntegerv(GL_MATRIX_MODE, &r); return r; }
inline GLmatrix getModelViewMatrix () { GLmatrix r; glGetFloatv(GL_MODELVIEW_MATRIX, r.data()); return r; }
inline GLuint glGenTexture () { GLuint r; glGenTextures(1, &r); return r; }
inline void glScalef (GLfloat const s) { glScalef(s, s, s); }

class glMatrixModeSwitch : boost::noncopyable
{
  GLenum const from;

  public:

    glMatrixModeSwitch (GLenum const to)
      : from (glGetMatrixMode())
    {
      glMatrixMode(to);
    }

    ~glMatrixModeSwitch () { glMatrixMode(from); }
};

inline GLmatrix xRotationMatrix (double const a)
{
  double const b [] = { 1, 0, 0, 0,
                        0, std::cos(a), std::sin(a), 0,
                        0, -std::sin(a),  std::cos(a), 0,
                        0, 0, 0, 1 };
  GLmatrix r;
  std::copy(b, b + 16, r.data());
  return r;
}

inline GLmatrix yRotationMatrix (double const a)
{
  double const b [] = { std::cos(a), 0, -std::sin(a), 0,
                        0, 1, 0, 0,
                        std::sin(a), 0, std::cos(a), 0,
                        0, 0, 0, 1 };
  GLmatrix r;
  std::copy(b, b + 16, r.data());
  return r;
}

inline GLmatrix identityMatrix ()
{
  GLmatrix r;
  for (int i = 0; i != 4; ++i)
  for (int j = 0; j != 4; ++j)
    r(i, j) = (i == j ? 1 : 0);
  return r;
}

struct glStackedMatrix : boost::noncopyable
{
  glStackedMatrix () { glPushMatrix(); }
  ~glStackedMatrix () { glPopMatrix(); }
};

#define GL_MATRIX(a00, a04, a08, a12, \
                  a01, a05, a09, a13, \
                  a02, a06, a10, a14, \
                  a03, a07, a11, a15) { a00, a01, a02, a03, \
                                        a04, a05, a06, a07, \
                                        a08, a09, a10, a11, \
                                        a12, a13, a14, a15 }

class Texture : boost::noncopyable
{
  GLuint const handle;
  public:
    Texture () : handle (glGenTexture()) {}
    ~Texture () { glDeleteTextures(1, &handle); }
    void bind () { glBindTexture(GL_TEXTURE_2D, handle); }

  /*

  Todo: Rewrite this, because textures are now used differently in FieldDrawers.

  This wrapper is noncopyable and immutable, which is sufficient for M3D's needs. This is very fortunate because making it copyable and/or mutable turns out to be very tricky. The main problems are:
  a) It's virtually impossible to mold OpenGL's curious API into a more sensible OO API because the former uses the "global" context to determine which object/resource "receives" a function call.
  b) It's virtually impossible to guard const-ness because our grasp over the resource is simply not strong enough, and without const-ness a lot of other things (such as copying and passing it around sensibly) break down, too.

  (All this assumes that we want to provide shared-resource semantics by default.)

  One possible design approach that may be able to cope with these problem to some degree is refrain from trying to implement a class T corresponding directly to the texture resource, but rather implement a class with shared_ptr<T> semantics. Thus, when we use this class in the program, the texture resource will implicitly be considered nonconst at all times. That is, we assume that there's no such thing as a constant texture resource (since we wouldn't be able to enforce it anyway).

  We may need to do this at some point in the future. However, for now the current approach suffices, even though one of its consequences is that textures can only be "constructed"/"acquired" once at their final owner, and can not be moved around afterwards.

  */
};

#endif // header guard
