
#ifndef UTIL_OVERLOADTAG_HPP
#define UTIL_OVERLOADTAG_HPP

/*

Consider a situation in which a library needs some user-supplied function associated with a type in order for that type to be used with a library service. One way to associate the function with the type is to make it an explicit specialization of a service-defined function template:

  // library code:
  template <typename> void f ();
  template <typename T> void service { ... f<T>(); ... }

  // user code:
  struct S { ... };
  template <> void f<S> () { ... }

This approach has at least the following two drawbacks:

  - f cannot be partially specialized, so users can only specialize f for concrete types.
  - users must define their specialization in the predetermined namespace rather than in S's namespace.

Another approach is to make the function a static member function of the type:

  // library code:
  template <typename T> void service { ... T::f(); ... }

  // user code:
  struct S { static void f () { ... } ... };

This approach has the drawback that f cannot be defined for built-in types or types not under user control.

Another approach is:

  // library code
  template <typename T> void service { ... f((T*)0); ... }

  // user code:
  struct S { ... };
  void f (T *) { ... }

This approach has none of the drawbacks mentioned above: f can be a template to achieve "partial specialization", f can be defined in S's namespace because of ADL, and f can be defined for any type. The extra parameter should be optimized away by modern compilers, so performance isn't affected.
In order to make this approach a bit slicker, we define:

*/

struct OverloadTag_base {};
template <typename> struct OverloadTag : OverloadTag_base {};

/*

Now, we can use:

  // library code
  template <typename T> void service { ... f(OverloadTag<T>()); ... }

  // user code:
  struct S { ... };
  void f (OverloadTag<T>) { ... }

Replacing the pointer with OverloadTag is more concise, makes the intent clearer, and is potentially more efficient.

*/

#endif // header guard
