
/*

Settings are usually just a bunch of simple values. Intuitively one might expect that these would be best represented as a POD-struct type. However, often a certain amount of validation can be done on the values. This validation must be performed every time new values enter the system (through a user configuring the application through a GUI, by reading configuration files, etc.). While it is possible to simply initiate the validation every time this happens, doing so manually is error prone. What we'd rather have is a non-POD type that validates the values upon initialization and modification.

Validated<T> wraps an object of type T. It validates its wrapped object on initialization and guarantees validity of its wrapped object at all times. It performs validation by evaluating validate(x), where x is the wrapped object of type T. validate() should throw a ::ValidationError exception if the options aren't valid. This ::ValidationError will be intercepted by Validated<T>, which will translate it to a Validated<T>::ValidationError, thereby adding type information.

Validated<T> is CopyConstructible iff T is CopyConstructible.

Note that the validation performed is determined only by type, so validated<> can't be used to implement things like even_integer and odd_integer, since those would be of the same type (assuming they are both typedefs for int).

Validated forwards operator<< and operator== to base_type.

Important: make sure your validate() takes a base_type parameter, not the Validated<T>. Though it will compile, it will cause an endless recursion (because in order to be able to call it, a new Validated<T> object will be created, but during the construction of that object validate() is called, etc etc etc). Note that this sneaky problem could be avoided if validated's constructor was explicit, but we really want implicit conversion from base_type to validated.

*/

#ifndef UTIL_VALIDATED_HPP
#define UTIL_VALIDATED_HPP

#include <stdexcept>
#include <string>
#include <iosfwd>

struct ValidationError : std::runtime_error
{
  explicit ValidationError (std::string const & s)
    : std::runtime_error(s)
  {}
};

template <typename T>
class Validated
{
  T t;

  public:

    typedef T base_type;

    class ValidationError : public ::ValidationError
    {
      friend class Validated;

      explicit ValidationError (std::string const & s)
        : ::ValidationError (s)
      {}
    };

    Validated (T const & t_) try : t (t_) { validate(t); }
    catch (::ValidationError const & e) { throw ValidationError(e.what()); } // adds type info

    T const * operator-> () const { return &t; }
    T const & operator* () const { return t; }
    T const & base () const { return t; }
};

// forward operator<<, operator==, to base_type

template <typename Char, typename Traits, typename T>
std::basic_ostream<Char, Traits> & operator<< (std::basic_ostream<Char, Traits> & o,
                                               Validated<T> const & t)
{
  return o << t.base();
}

template <typename T>
bool operator== (Validated<T> const & l, Validated<T> const & r)
{
  return l.base() == r.base();
}

#endif // header guard
