
// The standard guarantees that std::fstream throws an ios_base::failure if opening fails and the proper exception mask is set. However, it does not specify the what() value of that exception, and in practice implementations don't always provide a descriptive message. Hence, it cannot be relied on. Fortunately, the standard specifies that std::basic_filebuf::open behaves as if std::fopen was called, which in turn implies that errno will be set. Furthermore, std::strerror is guaranteed to return an error message corresponding to the error code in errno.

#ifndef UTIL_BFSTREAM_HPP
#define UTIL_BFSTREAM_HPP

#include <ios>
#include <string>
#include <exception>

#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/path.hpp>

class file_open_error : public std::exception
{
  boost::filesystem::path path_;
  std::ios_base::openmode mode_;
  int errcode_;
  std::string what_;

  public:

    file_open_error (boost::filesystem::path const &, std::ios_base::openmode, int errcode);

    boost::filesystem::path const & path () const { return path_; }
    std::ios_base::openmode mode () const { return mode_; }
    int errcode () const { return errcode_; }

    char const * what () const throw () { return what_.c_str(); }

    virtual ~file_open_error () throw () {}
};

// bfstream differs from boost::filesystem::fstream in that constructing opens in binary mode, can throw a file_open_error exception, and sets an exception mask.

// We'd like to hide open(), rdbuf(), and exceptions(), but we can't.

struct bfstream : boost::filesystem::fstream
{
  explicit bfstream (boost::filesystem::path const &, std::ios_base::openmode);
};

// bifstream and bofstream are merely shortcuts for bfstream's with mode ios_base::in and ios_base::out

struct bifstream : bfstream { explicit bifstream (boost::filesystem::path const &); };
struct bofstream : bfstream { explicit bofstream (boost::filesystem::path const &); };

#endif // header guard
