
#ifndef POD_SERIALIZATION_HPP
#define POD_SERIALIZATION_HPP

#include <boost/noncopyable.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/split_free.hpp>
#include <boost/utility/enable_if.hpp>

#include "PodIteratorBase.hpp"

#include <string>
#include <utility>
#include <stdexcept>

#include "serialization-symbols.hpp"

namespace symbolic_dispatch
{
  template <typename Archive, typename T, typename Pod, T (Pod:: * M)>
  typename boost::disable_if<is_symbolically_serialized<T, Pod, M>, void>::type
    save_pod_member (Archive & a, char const * const name, T const & t)
  {
    a << boost::serialization::make_nvp(name, t);
  }

  template <typename Archive, typename T, typename Pod, T (Pod:: * M)>
  typename boost::enable_if<is_symbolically_serialized<T, Pod, M>, void>::type
    save_pod_member (Archive & a, char const * const name, T const & t)
  {
    a << boost::serialization::make_nvp(name, to_symbol<T, Pod, M>(t));
  }

  template <typename Archive, typename T, typename Pod, T (Pod:: * M)>
  typename boost::disable_if<is_symbolically_serialized<T, Pod, M>, void>::type
    load_pod_member (Archive & a, char const * const name, T & t)
  {
    a >> boost::serialization::make_nvp(name, t);
  }

  template <typename Archive, typename T, typename Pod, T (Pod:: * M)>
  typename boost::enable_if<is_symbolically_serialized<T, Pod, M>, void>::type
    load_pod_member (Archive & a, char const * const name, T & t)
  {
    std::string q;
    a >> boost::serialization::make_nvp(name, q);
    t = from_symbol<T, Pod, M>(q);
  }
}

template <typename Pod, typename Archive>
class PodSaver : public PodIteratorBase<PodSaver<Pod, Archive>, Pod>, boost::noncopyable
{
  Archive & archive;
  Pod const & pod;

  public:

    template <typename T, T (Pod:: * M)>
    bool process (char const name [])
    {
      symbolic_dispatch::save_pod_member<Archive, T, Pod, M>(archive, name, pod.*M);
      return true;
    }

    PodSaver (Archive & a, Pod const & p) : archive (a), pod (p) {}
};

template <typename Pod, typename Archive>
class PodLoader : public PodIteratorBase<PodLoader<Pod, Archive>, Pod>, boost::noncopyable
{
  Archive & archive;
  Pod & pod;

  public:

    template <typename T, T (Pod:: * M)>
    bool process (char const name [])
    {
      symbolic_dispatch::load_pod_member<Archive, T, Pod, M>(archive, name, pod.*M);
      return true;
    }

    PodLoader (Archive & a, Pod & p) : archive (a), pod (p) {}
};


// POD_DEFINE_SERIALIZATION _must_ be used in global namespace scope.

// Todo: Why can't we use the SPLIT_FREE macro?

#define POD_DEFINE_SERIALIZATION(t) \
  namespace boost { \
    namespace serialization { \
      template <typename Archive> \
      void save (Archive & a, t const & s, unsigned int /* version */) \
      { \
        PodSaver<t, Archive> f (a, s); \
        pod_members(f, OverloadTag<t>()); \
      } \
      template <typename Archive> \
      void load (Archive & a, t & s, unsigned int /* version */) \
      { \
        PodLoader<t, Archive> f (a, s); \
        pod_members(f, OverloadTag<t>()); \
      } \
      template <typename Archive> \
      void serialize (Archive & a, t & s, unsigned int const file_version) \
      { \
          split_free(a, s, file_version); \
      } \
  } }

#endif // header guard
