
/*

In the GoF visitor pattern, the visitor side dispatch consists of a single plain virtual function call, initiated by the visitable's accept function. This is possible because the root visitor interface declares an exhaustive listing of virtual functions for all visitable types. While this simplicity is a significant advantage, the exhaustive listing of visitable types is its achilles heel. The approach implemented in this file adds a dispatch step consisting of a function template call on the visitor side just before the virtual call to the final visit() function that attempts to dynamic cast to an intermediary type in the visitor hierarchy that introduces the virtual visit function for the given visitable type. This approach has weaknesses, but no longer requires an exhaustice listing of visitable types in the root of visitor hierarchy. See the testing section at the end of this file for an extensive example.

Todo: Describe problems this approach has when dealing with non-flat visitable hierachies.

Todo: Perhaps it's a good idea to template parameterize the whole framework on an "Interface" which would be the fallback type for visit_other().

*/

#ifndef UTIL_VISITATION_HPP
#define UTIL_VISITATION_HPP

#include "VirtuallyDestructed.hpp"

template <typename T = void>
struct Visitor;

template <>
struct Visitor<void> : VirtuallyDestructed
{
  template <typename T>
  void visit (T & o)
  {
    if (Visitor<T> * const p = dynamic_cast<Visitor<T> *>(this)) p->visit(o);
    else this->visit_other();
  }

  template <typename T>
  void visit (T & o) const
  {
    if (Visitor<T> const * const p = dynamic_cast<Visitor<T> const *>(this)) p->visit(o);
    else this->visit_other();
  }

  virtual void visit_other () { Visitor const & t = *this; t.visit_other(); }; // by default just call the const version
  virtual void visit_other () const {}
};

template <typename T>
struct Visitor : virtual Visitor<>
{
  virtual void visit (T & u) { Visitor const & t = *this; t.visit(u); } // by default just call the const version
  virtual void visit (T &) const {}
};

template <typename Derived = void>
struct Visitable;

template <>
struct Visitable<void>
{
  virtual void accept (Visitor<>       &)       = 0;
  virtual void accept (Visitor<> const &)       = 0;
  virtual void accept (Visitor<>       &) const = 0;
  virtual void accept (Visitor<> const &) const = 0;
  virtual ~Visitable () {}
};

template <typename Derived>
class Visitable : public virtual Visitable<>
{
  void accept (Visitor<>       & v)       { v.visit(*static_cast<Derived *>(this)); }
  void accept (Visitor<> const & v)       { v.visit(*static_cast<Derived *>(this)); }
  void accept (Visitor<>       & v) const { v.visit(*static_cast<Derived const *>(this)); }
  void accept (Visitor<> const & v) const { v.visit(*static_cast<Derived const *>(this)); }
};

#endif // ifndef UTIL_VISITATION_HPP

#ifdef VISITATION_TEST

#include <iostream>
#include <vector>
#include <cassert>

struct Apple  : Visitable<Apple>  {};
struct Pear   : Visitable<Pear>   {};
struct Orange : Visitable<Orange> {};

struct MyFruitVisitor : Visitor<Apple      >, Visitor<Pear      >,
                        Visitor<Apple const>, Visitor<Pear const>
{
  mutable std::string s;

  void visit (Apple       &)       { s += 'a'; }
  void visit (Pear        &)       { s += 'b'; }
  void visit (Apple       &) const { s += 'c'; }
  void visit (Pear        &) const { s += 'd'; }
  void visit (Apple const &)       { s += 'e'; }
  void visit (Pear  const &)       { s += 'f'; }
  void visit (Apple const &) const { s += 'g'; }
  void visit (Pear  const &) const { s += 'h'; }

  void visit_other () const { s += 'i'; }
  void visit_other ()       { s += 'j'; }
};

int main ()
{
  Apple apple; Pear pear; Orange orange;

  typedef std::vector<Visitable<> *> V;
  typedef std::vector<Visitable<> const *> CV;

  V v;

  v.push_back(&apple);
  v.push_back(&pear);
  v.push_back(&orange);

  CV cv (v.begin(), v.end());

  MyFruitVisitor f;
  MyFruitVisitor const & cf = f;

  for ( V::iterator i =  v.begin(); i !=  v.end(); ++i) (*i)->accept( f);
  for ( V::iterator i =  v.begin(); i !=  v.end(); ++i) (*i)->accept(cf);
  for (CV::iterator i = cv.begin(); i != cv.end(); ++i) (*i)->accept( f);
  for (CV::iterator i = cv.begin(); i != cv.end(); ++i) (*i)->accept(cf);

  assert(f.s == "abjcdiefjghi");
}

#endif // ifdef VISITATION_TEST
